from html import escape as html_escape
from http import HTTPStatus

from rouver.types import BadArgumentsDict


def http_status_page(
    status: HTTPStatus,
    *,
    message: str = "",
    html_message: str = "",
    html_content: str = "",
) -> str:
    """Create an HTML error page for a given status code.

    A message and further content can optionally be provided.

    WARNING: The arguments `html_message` and `html_content` are not safe!
    Their content will be pasted into the page as is. Do not use with
    unsanitized data!
    """

    if message and html_message:
        raise ValueError("'message' and 'html_message' are mutually exclusive")

    if message:
        html_message = html_escape(message)

    paragraph = (
        "\n        <p>{}</p>".format(html_message) if html_message else ""
    )
    content = html_content + "\n" if html_content else ""
    return """<!DOCTYPE html>
<html>
    <head>
        <title>{0.value} &#x2014; {0.phrase}</title>
    </head>
    <body>
        <h1>{0.value} &#x2014; {0.phrase}</h1>{1}
{2}    </body>
</html>
""".format(
        status, paragraph, content
    )


def created_at_page(url: str) -> str:
    message = 'Created at <a href="{0}">{0}</a>.'.format(html_escape(url))
    return http_status_page(HTTPStatus.CREATED, html_message=message)


def temporary_redirect_page(url: str) -> str:
    message = 'Please see <a href="{0}">{0}</a>.'.format(html_escape(url))
    return http_status_page(
        HTTPStatus.TEMPORARY_REDIRECT, html_message=message
    )


def see_other_page(url: str) -> str:
    message = 'Please see <a href="{0}">{0}</a>.'.format(html_escape(url))
    return http_status_page(HTTPStatus.SEE_OTHER, html_message=message)


def bad_arguments_page(arguments: BadArgumentsDict) -> str:
    content = bad_arguments_list(arguments)
    return http_status_page(
        HTTPStatus.BAD_REQUEST,
        html_message="Invalid arguments:",
        html_content=content,
    )


def bad_arguments_list(arguments: BadArgumentsDict) -> str:
    if not arguments:
        return ""

    def format_item(name: str, error: str) -> str:
        return """    <li class="argument">
        <span class="argument-name">{name}</span>:
        <span class="error-message">{error}</span>
    </li>
""".format(
            name=html_escape(name), error=html_escape(error)
        )

    items = [format_item(k, arguments[k]) for k in sorted(arguments)]

    return '<ul class="bad-arguments">\n{}</ul>\n'.format("".join(items))



from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Mapping, Tuple

from typing_extensions import TypeAlias
from werkzeug.wrappers import Request

# (name, value)
Header: TypeAlias = Tuple[str, str]

WSGIEnvironment: TypeAlias = Dict[str, Any]

# (body) -> None
StartResponseReturnType: TypeAlias = Callable[[bytes], object]

# (status: str, headers: List[Headers], exc_info) -> response
StartResponse: TypeAlias = Callable[..., StartResponseReturnType]

WSGIResponse: TypeAlias = Iterable[bytes]

WSGIApplication: TypeAlias = Callable[
    [WSGIEnvironment, StartResponse], WSGIResponse
]

# (method, path, callback)
RouteDescription: TypeAlias = Tuple[str, str, WSGIApplication]

# (request, previous_args, path_part) -> result
RouteTemplateHandler: TypeAlias = Callable[
    [Request, Tuple[Any, ...], str], Any
]

BadArgumentsDict: TypeAlias = Mapping[str, str]



from urllib.parse import quote, urljoin

from werkzeug import Request


def absolute_url(request: Request, path: str) -> str:
    """
    Construct an absolute URL, using the request URL as base.

    Non-printable and non-ASCII characters in the path are encoded,
    but other characters, most notably slashes and percent signs are
    not encoded. Make sure to call urllib.parse.quote() on paths that
    can potentially contain such characters before passing them to
    absolute_url().
    """
    base_url: str = request.base_url
    return urljoin(base_url, quote(path, safe="/:?&$@,;+=%~"))


def rfc5987_encode(key: str, s: str) -> str:
    """Encode a string according to RFC 5987 if necessary."""
    if is_rfc2616_token(s):
        return f"{key}={s}"
    elif s.isascii():
        return f"{key}={rfc2616_quote(s)}"
    else:
        return f"{key}*=UTF-8''{quote(s)}"


RFC2616_SEPARATORS = '()<>@,;:\\"/[]?={} \t'


def is_rfc2616_token(s: str) -> bool:
    """Return whether `s` is a token according to RFC 2616."""
    for c in s:
        if ord(c) < 32 or ord(c) > 126:
            return False
        if c in RFC2616_SEPARATORS:
            return False
    return True


def rfc2616_quote(s: str) -> str:
    """Quote a string according to RFC 2616."""
    quoted = ""
    for c in s:
        if ord(c) < 32 or ord(c) > 126:
            quoted += f"\\{c}"
        elif c == '"':
            quoted += '\\"'
        else:
            quoted += c
    return f'"{quoted}"'



from __future__ import annotations

import logging
import re
from collections.abc import Iterable, Iterator, Mapping, Sequence
from enum import Enum
from http import HTTPStatus
from typing import Any, Tuple, cast
from urllib.parse import unquote

from werkzeug.exceptions import HTTPException, MethodNotAllowed, NotFound
from werkzeug.wrappers import Request

from rouver.exceptions import ArgumentsError
from rouver.html import bad_arguments_page, http_status_page
from rouver.response import respond_with_html
from rouver.types import (
    BadArgumentsDict,
    RouteDescription,
    RouteTemplateHandler,
    StartResponse,
    WSGIApplication,
    WSGIEnvironment,
)

LOGGER_NAME = "rouver"


class _TemplatePartType(Enum):
    STATIC = 1
    PATTERN = 2


def _split_path(s: str) -> list[str]:
    if not s:
        return []
    return s.split("/")


class Router:
    def __init__(self) -> None:
        self._handlers: list[_RouteHandler] = []
        self._sub_routers: list[_SubRouterHandler] = []
        self._template_handlers: dict[str, RouteTemplateHandler] = {}
        self.error_handling = True

    def __call__(
        self, environment: WSGIEnvironment, start_response: StartResponse
    ) -> Iterable[bytes]:
        try:
            return _dispatch(
                environment,
                start_response,
                self._handlers,
                self._sub_routers,
                self._template_handlers,
            )
        except Exception:
            if self.error_handling:
                logging.getLogger(LOGGER_NAME).exception(
                    "error while handling request"
                )
                return _respond_internal_server_error(start_response)
            else:
                raise

    def add_routes(self, routes: Sequence[RouteDescription]) -> None:
        self._handlers.extend(
            [_RouteHandler(r, self._template_handlers) for r in routes]
        )

    def add_template_handler(
        self, name: str, handler: RouteTemplateHandler
    ) -> None:
        self._template_handlers[name] = handler

    def add_sub_router(self, path: str, sub_router: WSGIApplication) -> None:
        """Handle a path using a sub-router.

        The following router will handle a route "/foo/sub":

        >>> def handle_sub(_, start_response):
        ...     start_response("200 OK", [])
        ...     return []
        >>> sub = Router()
        >>> sub.add_routes([("sub", "GET", handle_sub)])
        >>> router = Router()
        >>> router.add_sub_router("foo", sub)

        Sub routers can be Routers or any other WSGI application:

        >>> def sub_app(environ, start_response):
        ...     start_response("204 No Content", [])
        ...     return []
        >>> router = Router()
        >>> router.add_sub_router("foo", sub_app)

        Sub routers are matched after regular paths. This allows you to
        override selected paths in the super router:

        >>> def redirect(_, sr):
        ...     sr("307 Temporary Redirect",
        ...        [("Location", "https://www.example.com/")])
        ...     return []
        >>> sub = Router()
        >>> sub.add_routes([("sub", "GET", handle_sub)])
        >>> router = Router()
        >>> router.add_routes([("foo/bar", "GET", redirect)])
        >>> router.add_sub_router("foo", sub)
        """

        self._sub_routers.append(
            _SubRouterHandler(path, sub_router, self._template_handlers)
        )


_pattern_re = re.compile(r"^{(.*)}$")


def _parse_route_template_part(
    part: str, template_handlers: Mapping[str, RouteTemplateHandler]
) -> tuple[_TemplatePartType, str]:
    if part == "*":
        raise ValueError("wildcard not at end of path")
    m = _pattern_re.match(part)
    if m:
        name = m.group(1)
        if name not in template_handlers:
            msg = "unknown template path handler: '{}'".format(name)
            raise KeyError(msg) from None
        return _TemplatePartType.PATTERN, name
    else:
        return _TemplatePartType.STATIC, part


def _parse_path(
    path_string: str,
    template_handlers: Mapping[str, RouteTemplateHandler],
    *,
    allow_wildcard: bool = False,
) -> tuple[list[tuple[_TemplatePartType, str]], bool]:
    parts = _split_path(path_string)
    if allow_wildcard and parts and parts[-1] == "*":
        parts = parts[:-1]
        wildcard = True
    else:
        wildcard = False
    templates = [
        _parse_route_template_part(part, template_handlers) for part in parts
    ]
    return templates, wildcard


class _RouteHandler:
    def __init__(
        self,
        route: RouteDescription,
        template_handlers: Mapping[str, RouteTemplateHandler],
    ) -> None:
        self.path, self.wildcard = _parse_path(
            route[0], template_handlers, allow_wildcard=True
        )
        self.method = route[1]
        self.handler = route[2]


class _SubRouterHandler:
    def __init__(
        self,
        path: str,
        router: WSGIApplication,
        template_handlers: Mapping[str, RouteTemplateHandler],
    ) -> None:
        self.path, _ = _parse_path(path, template_handlers)
        self.router = router


def _dispatch(
    environment: WSGIEnvironment,
    start_response: StartResponse,
    handlers: Sequence[_RouteHandler],
    sub_routers: Sequence[_SubRouterHandler],
    template_handlers: Mapping[str, RouteTemplateHandler],
) -> Iterable[bytes]:
    request = Request(environment)
    path = _split_path(request.path[1:])
    arguments = _RouteArguments(request, template_handlers)

    def find_route_and_call_handler() -> Iterable[bytes]:
        try:
            matcher = find_route()
        except NotFound:
            pass
        except MethodNotAllowed as exc:
            return _respond_method_not_allowed(
                start_response, request.method, exc.valid_methods
            )
        else:
            return call_handler(matcher)

        try:
            sub_matcher = find_sub_router()
        except NotFound:
            pass
        else:
            return call_sub_router(sub_matcher)

        return _respond_not_found(environment, start_response)

    def find_route() -> _RouteMatcher:
        matchers = [_RouteMatcher(h, path, arguments) for h in handlers]

        matching_paths = [m for m in matchers if m.matches]
        if not matching_paths:
            raise NotFound()

        matching_routes = [
            m for m in matching_paths if m.method == request.method
        ]
        if not matching_routes:
            valid_methods = sorted(set(m.method for m in matching_paths))
            raise MethodNotAllowed(valid_methods)

        assert len(matching_routes) >= 1
        # Workaround for https://github.com/python/mypy/issues/4345
        assert isinstance(matching_routes[0], _RouteMatcher)
        return matching_routes[0]

    def call_handler(matcher: _RouteMatcher) -> Iterable[bytes]:
        try:
            return matcher.call(environment, start_response)
        except ArgumentsError as exc:
            return _respond_arguments_error(start_response, exc.arguments)
        except HTTPException as exc:
            return _respond_http_exception(start_response, exc)

    def find_sub_router() -> _SubRouterMatcher:
        matchers = [
            _SubRouterMatcher(sub_r, path, arguments) for sub_r in sub_routers
        ]
        for m in matchers:
            if m.matches:
                return m
        raise NotFound()

    def call_sub_router(matcher: _SubRouterMatcher) -> Iterable[bytes]:
        old_path = environment.get("PATH_INFO", "")
        remaining_path = matcher.remaining_path
        assert isinstance(old_path, str)
        new_environ = environment.copy()
        new_environ["rouver.original_path_info"] = old_path
        new_environ["SCRIPT_NAME"] = (
            environment.get("SCRIPT_NAME", "")
            + old_path[: -len(remaining_path)]
        )
        new_environ["PATH_INFO"] = remaining_path.encode("utf-8").decode(
            "latin-1"
        )
        return matcher.call(new_environ, start_response)

    return find_route_and_call_handler()


class _RouteArguments:
    def __init__(
        self,
        request: Request,
        template_handlers: Mapping[str, RouteTemplateHandler],
    ) -> None:
        self._request = request
        self._handlers = template_handlers
        self._cache: dict[tuple[str, str], Any] = {}

    def parse_argument(
        self, paths: Tuple[Any, ...], name: str, path: str
    ) -> Any:
        key = name, path
        if key not in self._cache:
            handler = self._handlers[name]
            self._cache[key] = handler(self._request, paths, path)
        return self._cache[key]


class _MatcherBase:
    def __init__(
        self,
        match_path: Sequence[tuple[_TemplatePartType, str]],
        request_path: Sequence[str],
        arguments: _RouteArguments,
        *,
        match_full_path: bool = False,
    ) -> None:
        if request_path and request_path[-1] == "":
            request_path = request_path[:-1]
            self._trailing_slash = True
        else:
            self._trailing_slash = False
        self._match_path = match_path
        self._request_path = request_path
        self._arguments = arguments
        self._match_full_path = match_full_path
        self.matches, self.path_args = self._check_and_parse()

    def _check_and_parse(self) -> tuple[bool, list[Any]]:
        if self._path_length_matches():
            return self._parse()
        else:
            return False, []

    def _path_length_matches(self) -> bool:
        if self._match_full_path:
            return len(self._match_path) == len(self._request_path)
        else:
            return len(self._match_path) <= len(self._request_path)

    def _parse(self) -> tuple[bool, list[Any]]:
        path_args: list[Any] = []
        for tmpl_part, path_part in self._path_compare_iter():
            try:
                decoded = unquote(path_part, errors="strict")
            except UnicodeDecodeError:
                return False, []
            tmpl_type, text = tmpl_part
            if tmpl_type == _TemplatePartType.STATIC:
                if text != decoded:
                    return False, []
            elif tmpl_type == _TemplatePartType.PATTERN:
                try:
                    arg = self._arguments.parse_argument(
                        tuple(path_args), text, decoded
                    )
                except ValueError:
                    return False, []
                path_args.append(arg)
            else:
                raise AssertionError("unhandled template type")
        return True, path_args

    def _path_compare_iter(
        self,
    ) -> Iterator[tuple[tuple[_TemplatePartType, str], str]]:
        return zip(self._match_path, self._request_path)

    @property
    def remaining_path(self) -> str:
        remaining_path = list(self._request_path[len(self._match_path) :])
        path = "/".join([""] + remaining_path)
        if self._trailing_slash:
            path += "/"
        return path


class _RouteMatcher(_MatcherBase):
    def __init__(
        self,
        handler: _RouteHandler,
        path: Sequence[str],
        arguments: _RouteArguments,
    ) -> None:
        super().__init__(
            handler.path, path, arguments, match_full_path=not handler.wildcard
        )
        self.method = handler.method
        self._handler = handler.handler

    def call(
        self, environ: WSGIEnvironment, start_response: StartResponse
    ) -> Iterable[bytes]:
        environ["rouver.path_args"] = self.path_args
        environ["rouver.wildcard_path"] = self.remaining_path
        return self._handler(environ, start_response)


class _SubRouterMatcher(_MatcherBase):
    def __init__(
        self,
        handler: _SubRouterHandler,
        path: Sequence[str],
        arguments: _RouteArguments,
    ) -> None:
        super().__init__(handler.path, path, arguments)
        self._router = handler.router

    def call(
        self, environ: WSGIEnvironment, start_response: StartResponse
    ) -> Iterable[bytes]:
        return self._router(environ, start_response)


def _respond_not_found(
    environment: WSGIEnvironment, start_response: StartResponse
) -> Iterable[bytes]:
    path = cast(str, environment.get("PATH_INFO", ""))
    message = "Path '{}' not found.".format(path)
    page = http_status_page(HTTPStatus.NOT_FOUND, message=message)
    return respond_with_html(start_response, page, status=HTTPStatus.NOT_FOUND)


def _respond_method_not_allowed(
    start_response: StartResponse,
    method: str,
    allowed_methods: Iterable[str] | None,
) -> Iterable[bytes]:
    method_string = " or ".join(allowed_methods or [])
    message = "Method '{}' not allowed. Please try {}.".format(
        method, method_string
    )
    html = http_status_page(HTTPStatus.METHOD_NOT_ALLOWED, message=message)
    return respond_with_html(
        start_response,
        html,
        status=HTTPStatus.METHOD_NOT_ALLOWED,
        extra_headers=[("Allow", ", ".join(allowed_methods or []))],
    )


def _respond_internal_server_error(
    start_response: StartResponse,
) -> Iterable[bytes]:
    html = http_status_page(
        HTTPStatus.INTERNAL_SERVER_ERROR, message="Internal server error."
    )
    return respond_with_html(
        start_response, html, status=HTTPStatus.INTERNAL_SERVER_ERROR
    )


def _respond_http_exception(
    start_response: StartResponse, exception: HTTPException
) -> Iterable[bytes]:
    assert exception.code is not None
    status = HTTPStatus(exception.code)
    html = http_status_page(status, message=exception.description or "")
    headers = [
        h for h in exception.get_headers() if h[0].lower() != "content-type"
    ]
    return respond_with_html(
        start_response,
        html,
        status=status,
        extra_headers=headers,
    )


def _respond_arguments_error(
    start_response: StartResponse, arguments: BadArgumentsDict
) -> Iterable[bytes]:
    html = bad_arguments_page(arguments)
    return respond_with_html(
        start_response, html, status=HTTPStatus.BAD_REQUEST
    )



from __future__ import annotations

import collections.abc
from collections.abc import Iterable, Iterator, Sequence
from http import HTTPStatus
from json import JSONDecodeError, loads as json_loads
from typing import Any, Protocol, cast
from urllib.parse import unquote

from werkzeug.exceptions import UnsupportedMediaType
from werkzeug.wrappers import Request

from rouver.args import ArgumentDict, ArgumentParser, ArgumentTemplate
from rouver.response import (
    created_as_json,
    created_at,
    respond,
    respond_with_content,
    respond_with_html,
    respond_with_json,
    see_other,
    temporary_redirect,
)
from rouver.types import Header, StartResponse, WSGIEnvironment


class _Closeable(Protocol):
    def close(self) -> object:
        ...


class RouteHandlerBase(collections.abc.Iterable):
    """Base class for rouver route handlers.

    Sub-classes of RouteHandlerBase can act as route handlers. They provide
    convenient, opaque access to several rouver services.

    Implementations must implement the prepare_response() method.

    >>> from rouver.router import Router
    >>> class MyRouteHandler(RouteHandlerBase):
    ...     def prepare_response(self):
    ...         return self.respond_with_html("<div>Hello World!</div>")

    >>> class MyRouter(Router):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.add_routes([
    ...             ("my-route", "GET", MyRouteHandler),
    ...         ])
    """

    def __init__(
        self, environ: WSGIEnvironment, start_response: StartResponse
    ) -> None:
        self.request = Request(environ)
        self.start_response = start_response
        self._argument_parser: ArgumentParser | None = None
        self._response = self.prepare_response()

    def __iter__(self) -> Iterator[bytes]:
        return iter(self._response)

    def close(self) -> None:
        if hasattr(self._response, "close"):
            cast(_Closeable, self._response).close()

    @property
    def _charset(self) -> str:
        return cast(str, self.request.mimetype_params.get("charset", "utf-8"))

    @property
    def path_args(self) -> list[Any]:
        path_args = self.request.environ.get("rouver.path_args")
        if not isinstance(path_args, list):
            return []
        return path_args

    @property
    def wildcard_path(self) -> str:
        path = self.request.environ.get("rouver.wildcard_path")
        if not isinstance(path, str):
            return ""
        return unquote(path)

    def prepare_response(self) -> Iterable[bytes]:
        raise NotImplementedError()

    def parse_args(
        self,
        argument_template: Sequence[ArgumentTemplate],
        *,
        exhaustive: bool = False,
    ) -> ArgumentDict:
        if self._argument_parser is None:
            environ = self.request.environ
            self._argument_parser = ArgumentParser(environ)
        return self._argument_parser.parse_args(
            argument_template, exhaustive=exhaustive
        )

    def parse_json_request(self) -> Any:
        """Parse the request body as JSON.

        Raise UnsupportedMediaType if the content type does not indicate
        a JSON request or if it contains invalid JSON.
        """

        if self.request.mimetype != "application/json":
            raise UnsupportedMediaType()
        try:
            j = self.request.data.decode(self._charset)
            return json_loads(j)
        except (LookupError, JSONDecodeError) as exc:
            raise UnsupportedMediaType(str(exc)) from exc

    def respond(
        self,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        content_type: str | None = None,
        extra_headers: Sequence[Header] = [],
    ) -> Iterable[bytes]:
        return respond(
            self.start_response,
            status=status,
            content_type=content_type,
            extra_headers=extra_headers,
        )

    def respond_with_content(
        self,
        content: bytes,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        content_type: str = "application/octet-stream",
        extra_headers: Sequence[Header] = [],
    ) -> Iterable[bytes]:
        """Prepare an WSGI response.

        >>> def handler(start_response, request):
        ...     return respond_with_content(start_response, b"content")

        The response will include a Content-Type and a Content-Length header.
        """
        return respond_with_content(
            self.start_response,
            content,
            status=status,
            content_type=content_type,
            extra_headers=extra_headers,
        )

    def respond_with_json(
        self,
        json: str | bytes | Any,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        extra_headers: Sequence[Header] = [],
    ) -> Iterable[bytes]:
        return respond_with_json(
            self.start_response,
            json,
            status=status,
            extra_headers=extra_headers,
        )

    def respond_with_html(
        self,
        html: str,
        *,
        status: HTTPStatus = HTTPStatus.OK,
        extra_headers: Sequence[Header] = [],
    ) -> Iterable[bytes]:
        return respond_with_html(
            self.start_response,
            html,
            status=status,
            extra_headers=extra_headers,
        )

    def created_at(self, url_part: str) -> Iterable[bytes]:
        return created_at(self.request, self.start_response, url_part)

    def created_as_json(
        self, url_part: str, json: str | bytes | Any
    ) -> Iterable[bytes]:
        return created_as_json(
            self.request, self.start_response, url_part, json
        )

    def temporary_redirect(self, url_part: str) -> Iterable[bytes]:
        return temporary_redirect(self.request, self.start_response, url_part)

    def see_other(self, url_part: str) -> Iterable[bytes]:
        return see_other(self.request, self.start_response, url_part)



from __future__ import annotations

import json
import re
import secrets
from collections.abc import Callable, Iterable, Sequence
from http import HTTPStatus
from io import BytesIO
from json import JSONDecodeError
from types import TracebackType
from typing import Any, Protocol, Union, cast
from urllib.parse import quote_plus, urlparse

from dectest import TestCase, before
from werkzeug.http import parse_options_header

from rouver.args import Multiplicity
from rouver.router import Router
from rouver.types import Header, WSGIApplication, WSGIEnvironment
from rouver.util import rfc5987_encode

_STATUS_RE = re.compile(r"^(\d\d\d) [ -~]+$")


class TestRequest:
    def __init__(self, method: str, path: str) -> None:
        self.method = method.upper()
        self.path = path
        self._body = b""
        self.error_stream = BytesIO()
        self.content_type: str | None = None
        self._extra_environ: WSGIEnvironment = {}
        self._extra_headers: list[tuple[str, str]] = []
        self._arguments: list[tuple[str, str]] = []
        self._file_arguments: list[tuple[str, bytes, str, str | None]] = []
        self._boundary: str | None = None
        self.script_name = ""

    @property
    def full_path(self) -> str:
        """Return the full path, without the query string.

        This is the concatenation of the script name and the path.
        """
        return self.script_name + self.path

    @property
    def body(self) -> bytes:
        return self._body

    @body.setter
    def body(self, body: bytes) -> None:
        if not isinstance(body, bytes):
            raise TypeError("body must be bytes")
        if body != b"" and self.method == "GET":
            raise ValueError("GET requests can not have a body")
        if self._arguments:
            raise ValueError(
                "setting arguments and a body is mutually exclusive"
            )
        self._body = body

    def set_json_request(
        self, body: str | bytes | dict[str, Any] | list[Any]
    ) -> None:
        """Send JSON data.

        body must either be UTF-8-encoded bytes, a string, or a dict or
        list containing values that can be encoded as JSON.

        The Content-Type header of the request will be set to
        "application/json; charset=utf-8".

        Sending JSON data will not work with GET requests.
        """
        self.content_type = "application/json; charset=utf-8"
        if isinstance(body, str):
            self.body = body.encode("utf-8")
        elif isinstance(body, bytes):
            self.body = body
        elif isinstance(body, (dict, list)):
            self.body = json.dumps(body).encode("ascii")
        else:
            raise TypeError("body must be one of str, bytes, list, or dict")

    def set_env_var(self, name: str, value: Any) -> None:
        self._extra_environ[name] = value

    def set_header(self, name: str, value: str) -> None:
        if name.lower() == "content-type":
            self.content_type = value
        else:
            self._extra_headers.append((name, value))

    def add_argument(self, name: str, value: str | Iterable[str]) -> None:
        """Add a CGI argument to this request.

        For GET and HEAD requests, this will add a query string to the
        URL, for other requests it will include the arguments in the
        request body.
        """
        if self.body != b"":
            raise ValueError(
                "setting arguments and a body is mutually exclusive"
            )
        values = [value] if isinstance(value, str) else list(value)
        for v in values:
            self._arguments.append((name, v))

    def add_file_argument(
        self,
        name: str,
        content: bytes,
        content_type: str,
        *,
        filename: str | None = None,
    ) -> None:
        """Add a file CGI argument to this request.

        This is not available for GET and HEAD requests. In other requests
        it will force a multipart request body.
        """
        if self.method in ["GET", "HEAD"]:
            raise ValueError(
                "file arguments not supported in GET and HEAD requests"
            )
        if self.body != b"":
            raise ValueError(
                "setting arguments and a body is mutually exclusive"
            )
        self._file_arguments.append((name, content, content_type, filename))

    def clear_arguments(self) -> None:
        self._arguments = []

    def to_environment(self) -> WSGIEnvironment:
        env = {
            "REQUEST_METHOD": self.method,
            "PATH_INFO": self.path,
            "SCRIPT_NAME": self.script_name,
            "SERVER_NAME": "www.example.com",
            "SERVER_PORT": "80",
            "SERVER_PROTOCOL": "HTTP/1.1",
            "REMOTE_ADDR": "127.0.0.1",
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": "http",
            "wsgi.errors": self.error_stream,
            "wsgi.multithread": False,
            "wsgi.multiprocess": False,
            "wsgi.run_once": True,
        }
        for header, value in self._extra_headers:
            env["HTTP_" + header.upper().replace("-", "_")] = value
        body = self._body
        if self._file_arguments:
            assert body == b""
            body = self._build_multipart_body()
        elif self._arguments:
            if self.method == "GET":
                env["QUERY_STRING"] = self._build_query_string()
            else:
                assert body == b""
                body = self._build_query_string().encode("ascii")
        env["wsgi.input"] = BytesIO(body)
        if body != b"":
            env["CONTENT_LENGTH"] = str(len(body))
        content_type = self._determine_content_type()
        if content_type is not None:
            env["CONTENT_TYPE"] = content_type
        env.update(self._extra_environ)
        return env

    def _determine_content_type(self) -> str | None:
        if self.content_type is not None:
            return self.content_type
        elif self._file_arguments:
            boundary = self._ensure_boundary()
            return f"multipart/form-data; boundary={boundary}"
        elif self.method not in ["GET", "HEAD"] and self._arguments:
            return "application/x-www-form-urlencoded"
        else:
            return None

    def _build_query_string(self) -> str:
        parts: list[str] = []
        for name, value in self._arguments:
            parts.append("{}={}".format(quote_plus(name), quote_plus(value)))
        return "&".join(parts)

    def _build_multipart_body(self) -> bytes:
        boundary = self._ensure_boundary()
        body = b""
        for name, value in self._arguments:
            body += b"--" + boundary.encode("ascii") + b"\r\n"
            body += (
                b'Content-Disposition: form-data; name="'
                + quote_plus(name).encode("ascii")
                + b'"\r\n\r\n'
            )
            body += value.encode("utf-8")
            body += b"\r\n"
        for name, content, content_type, filename in self._file_arguments:
            body += b"--" + boundary.encode("ascii") + b"\r\n"
            body += (
                b'Content-Disposition: form-data; name="'
                + quote_plus(name).encode("ascii")
            ) + b'"; '
            if filename:
                body += rfc5987_encode("filename", filename).encode("ascii")
            else:
                body += b'filename=""'
            body += b"\r\n"
            body += b"Content-Type: " + content_type.encode("ascii") + b"\r\n"
            body += (
                b"Content-Length: "
                + str(len(content)).encode("ascii")
                + b"\r\n\r\n"
            )
            body += content + b"\r\n"
        body += b"--" + boundary.encode("ascii") + b"--\r\n"
        return body

    def _ensure_boundary(self) -> str:
        if self._boundary is None:
            self._boundary = secrets.token_hex()
        return self._boundary


def create_request(method: str, path: str) -> TestRequest:
    return TestRequest(method, path)


class TestResponse:
    def __init__(self, status_line: str, headers: list[Header]) -> None:
        m = _STATUS_RE.match(status_line)
        if not m:
            raise ValueError("invalid status line")
        self.status_line = status_line
        self.status = HTTPStatus(int(m.group(1)))
        self._headers = headers
        self.body = b""

    def get_header_value(self, name: str) -> str:
        for n, v in self._headers:
            if n.lower() == name.lower():
                return v
        raise ValueError("header '{}' not in response".format(name))

    def parse_json_body(self) -> Any:
        """Return the response body as a JSON value.

        Raise an AssertionError if the response headers are incorrect or
        the JSON response is invalid.
        """
        self.assert_content_type(
            "application/json", charset=[None, "us-ascii", "utf-8"]
        )
        try:
            return json.loads(self.body.decode("utf-8"))
        except (UnicodeDecodeError, JSONDecodeError) as exc:
            raise AssertionError(str(exc)) from exc

    def assert_status(self, status: HTTPStatus) -> None:
        assert (
            status == self.status
        ), f"unexpected HTTP status: {status.value} != {self.status.value}"

    def assert_header_not_set(self, name: str) -> None:
        for n, v in self._headers:
            if n.lower() == name.lower():
                raise AssertionError(f"header '{name}' unexpectedly set")

    def assert_header_equal(self, name: str, expected_value: str) -> None:
        try:
            real_value = self.get_header_value(name)
        except ValueError:
            raise AssertionError("missing header '{}'".format(name))
        assert real_value == expected_value, (
            f"header value of '{name}' differs: "
            f"{real_value!r} != {expected_value!r}"
        )

    def _assert_location_response(
        self, expected_status: HTTPStatus, expected_location: str
    ) -> None:
        self.assert_status(expected_status)
        if ":" in expected_location:
            self.assert_header_equal("Location", expected_location)
        else:
            real_location = self.get_header_value("Location")
            parsed = urlparse(real_location)
            real_location = parsed.path
            if parsed.query:
                real_location += "?" + parsed.query
            if parsed.fragment:
                real_location += "#" + parsed.fragment
            assert real_location == expected_location, (
                f"unexpected location: {real_location!r} != "
                f"{expected_location!r}"
            )

    def assert_created_at(self, expected_location: str) -> None:
        """Assert a correct 201 Created response.

        The expected location can either be an absolute URL, or just the
        path portion.
        """
        self._assert_location_response(HTTPStatus.CREATED, expected_location)

    def assert_see_other(self, expected_location: str) -> None:
        """Assert a correct 303 See Other response.

        The expected location can either be an absolute URL, or just the
        path portion.
        """
        self._assert_location_response(HTTPStatus.SEE_OTHER, expected_location)

    def assert_temporary_redirect(self, expected_location: str) -> None:
        """Assert a correct 307 Temporary Redirect response.

        The expected location can either be an absolute URL, or just the
        path portion.
        """
        self._assert_location_response(
            HTTPStatus.TEMPORARY_REDIRECT, expected_location
        )

    def assert_content_type(
        self,
        content_type: str,
        *,
        charset: str | Sequence[str | None] | None = None,
    ) -> None:
        """Assert the response's Content-Type header.

        If the optional charset argument is given, compare the charset
        as well. This can be either a string or a sequence of strings. If
        the list includes the value None, the charset is optional.

        """
        try:
            value = self.get_header_value("Content-Type")
        except ValueError:
            raise AssertionError("missing header 'Content-Type'")
        type_, options = parse_options_header(value)
        assert (
            type_ == content_type
        ), f"unexpected content type: {type_!r} != {content_type!r}"
        if charset is not None:
            cs_list = [charset] if isinstance(charset, str) else charset
            try:
                got_charset = options["charset"]
            except KeyError:
                if None in cs_list:
                    return
                raise AssertionError("no charset in Content-Type header")
            assert got_charset in cs_list, (
                f"unexpected content type charset: "
                f"{got_charset!r} not in {cs_list!r}"
            )

    def assert_set_cookie(
        self,
        expected_name: str,
        expected_value: str,
        *,
        secure: bool | None = None,
        http_only: bool | None = None,
        max_age: int | None = None,
    ) -> None:
        def assert_flag(flag: bool | None, name_: str) -> None:
            if flag:
                if find_arg(name_) is None:
                    raise AssertionError(
                        f"Set-Cookie does not contain the '{name_}' flag"
                    )
            elif flag is not None and not flag:
                if find_arg(name_) is not None:
                    raise AssertionError(
                        f"Set-Cookie contains the '{name_}' flag unexpectedly"
                    )

        def find_arg(arg_name: str) -> str | None:
            for a in args:
                if a[0].lower() == arg_name.lower():
                    return a[1] if len(a) >= 2 else ""
            return None

        def expect_arg(arg_name: str) -> str:
            arg_value = find_arg(arg_name)
            if arg_value is None:
                raise AssertionError(
                    "Set-Cookie does not contain the '{}' argument".format(
                        name
                    )
                )
            return arg_value

        try:
            header_value = self.get_header_value("Set-Cookie")
        except ValueError:
            raise AssertionError("missing header 'Set-Cookie'")
        args = [s.strip().split("=", 1) for s in header_value.split(";")]
        if len(args[0]) < 2:
            raise AssertionError("invalid Set-Cookie header")
        name, value = args[0]
        assert (
            name == expected_name
        ), f"wrong cookie name, {name!r} != {expected_name!r}"
        assert (
            value == expected_value
        ), f"wrong cookie value, {value!r} != {expected_value!r}"
        assert_flag(secure, "Secure")
        assert_flag(http_only, "HttpOnly")
        if max_age is not None:
            age = expect_arg("Max-Age")
            try:
                int_age = int(age)
            except ValueError:
                raise AssertionError("Set-Cookie max-age is not numeric")
            assert (
                int_age == max_age
            ), f"wrong max-age: {int_age!r} != {max_age!r}"


class _Closeable(Protocol):
    def close(self) -> object:
        ...


def test_wsgi_app(app: WSGIApplication, request: TestRequest) -> TestResponse:
    output_written = False
    response: TestResponse | None = None

    def write(b: bytes) -> None:
        nonlocal output_written
        assert response is not None
        if b:
            response.body += b
            output_written = True

    def start_response(
        status: str,
        response_headers: list[Header],
        exc_info: tuple[type[BaseException], BaseException, TracebackType]
        | None = None,
    ) -> Callable[[bytes], Any]:
        nonlocal response
        if response and not exc_info:
            raise AssertionError("start_response called multiple times")
        if output_written:
            assert exc_info is not None
            raise exc_info[1].with_traceback(exc_info[2])
        response = TestResponse(status, response_headers)
        return write

    env = request.to_environment()
    result = app(env, start_response)
    if response is None:
        raise AssertionError("start_response() was not called")
    for data in result:
        response.body += data
    if hasattr(result, "close"):
        cast(_Closeable, result).close()
    return response


ArgumentToTest = Union[
    "tuple[str, Multiplicity, str]", "tuple[str, Multiplicity, str, str]"
]


def test_wsgi_arguments(
    app: WSGIApplication,
    request: TestRequest,
    arguments: Iterable[ArgumentToTest],
) -> None:
    def is_required(argument: ArgumentToTest) -> bool:
        return argument[1] in (
            Multiplicity.REQUIRED,
            Multiplicity.REQUIRED_ANY,
        )

    def setup_args(args: Iterable[ArgumentToTest]) -> None:
        request.clear_arguments()
        if not args and request.method != "GET":
            request.content_type = "application/x-www-form-urlencoded"
        for argument in args:
            assert len(argument) == 3 or len(argument) == 4
            name, multi, valid_value = argument[:3]
            request.add_argument(name, valid_value)

    def setup_required_args_except(argument_name: str) -> None:
        setup_args(
            a for a in arguments if is_required(a) and a[0] != argument_name
        )

    def call_expect_success() -> None:
        response = test_wsgi_app(app, request)
        if response.status == HTTPStatus.BAD_REQUEST:
            raise AssertionError(
                "Bad Request returned, although CGI arguments were correct"
            )
        elif response.status.value >= 400:
            raise AssertionError(
                f"status was: {response.status.value}, "
                f"but expected a successful result"
            )

    def call_expect_bad_request(message: str) -> None:
        response = test_wsgi_app(app, request)
        if response.status != HTTPStatus.BAD_REQUEST:
            raise AssertionError(message)

    def assert_success_required_arguments() -> None:
        setup_args(a for a in arguments if is_required(a))
        call_expect_success()

    def assert_success_all_arguments() -> None:
        setup_args(arguments)
        call_expect_success()

    def assert_failure_if_argument_required_but_missing(
        argument: ArgumentToTest,
    ) -> None:
        if not is_required(arg):
            return
        setup_required_args_except(argument[0])
        call_expect_bad_request(
            "Bad Request not returned, although required CGI argument "
            f"'{argument[0]}' was missing"
        )

    def assert_failure_if_argument_invalid(argument: ArgumentToTest) -> None:
        if len(argument) != 4:  # argument has no invalid value
            return
        name, _, __, invalid_value = argument  # type: ignore
        setup_required_args_except(name)
        request.add_argument(name, invalid_value)
        call_expect_bad_request(
            f"Bad Request not returned, although CGI argument '{name}' "
            f"had invalid value '{invalid_value}'"
        )

    assert_success_required_arguments()
    assert_success_all_arguments()
    for arg in arguments:
        assert_failure_if_argument_required_but_missing(arg)
    for arg in arguments:
        assert_failure_if_argument_invalid(arg)


class RouterTestCase(TestCase):
    @before
    def setup_router(self) -> None:
        self.router = self.create_router()
        self.router.error_handling = False

    def create_router(self) -> Router:
        raise NotImplementedError()

    def send_request(self, request: TestRequest) -> TestResponse:
        return test_wsgi_app(self.router, request)

    def assert_arguments(
        self, request: TestRequest, arguments: Iterable[ArgumentToTest]
    ) -> None:
        test_wsgi_arguments(self.router, request, arguments)


# pytest-friendly aliases
FakeRequest = TestRequest
FakeResponse = TestResponse
run_wsgi_test = test_wsgi_app



from __future__ import annotations

from collections.abc import Iterable, Sequence
from http import HTTPStatus
from json import dumps as dumps_json
from typing import Any

from werkzeug.wrappers import Request

from rouver.html import (
    created_at_page,
    see_other_page,
    temporary_redirect_page,
)
from rouver.status import status_line
from rouver.types import Header, StartResponse
from rouver.util import absolute_url


def _location_header(request: Request, url: str) -> Header:
    return "Location", absolute_url(request, url)


def respond(
    start_response: StartResponse,
    *,
    status: HTTPStatus = HTTPStatus.OK,
    content_type: str | None = None,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """Prepare an empty WSGI response.

    >>> def handler(start_response, request):
    ...     return respond(start_response, status=HTTPStatus.ACCEPTED)

    The return value can be ignored to return an arbitrary response body.

    >>> def handler(start_response, request):
    ...     respond(start_response, content_type="text/plain")
    ...     return [b"My Response"]
    """

    sl = status_line(status)
    all_headers = list(extra_headers)
    if content_type is not None:
        if any(h.lower() == "content-type" for (h, _) in extra_headers):
            raise ValueError("duplicate Content-Type header")
        all_headers += [("Content-Type", content_type)]
    start_response(sl, all_headers)
    return []


def respond_with_content(
    start_response: StartResponse,
    content: bytes,
    *,
    status: HTTPStatus = HTTPStatus.OK,
    content_type: str = "application/octet-stream",
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """Prepare an WSGI response.

    >>> def handler(start_response, request):
    ...     return respond_with_content(start_response, b"content")

    The response will include a Content-Type and a Content-Length header.

    """

    sl = status_line(status)
    headers = [
        ("Content-Type", content_type),
        ("Content-Length", str(len(content))),
    ] + list(extra_headers)
    start_response(sl, headers)
    return [content]


def respond_with_json(
    start_response: StartResponse,
    json: str | bytes | Any,
    *,
    status: HTTPStatus = HTTPStatus.OK,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """Prepare a JSON WSGI response.

    >>> def handler(start_response, request):
    ...     return respond_with_json(start_response, {"foo": "bar"})

    The JSON text to return can be supplied as a string, as an UTF-8-encoded
    bytestring, or as any object that can be serialized using json.dumps().

    The default response status of "200 OK" can be overridden with the
    "status" keyword argument.
    """

    if isinstance(json, bytes):
        encoded = json
    elif isinstance(json, str):
        encoded = json.encode("utf-8")
    else:
        encoded = dumps_json(json).encode("utf-8")

    return respond_with_content(
        start_response,
        encoded,
        status=status,
        content_type="application/json; charset=utf-8",
        extra_headers=extra_headers,
    )


def respond_with_html(
    start_response: StartResponse,
    html: str,
    *,
    status: HTTPStatus = HTTPStatus.OK,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """Prepare an HTML WSGI response.

    >>> def handler(start_response, request):
    ...     return respond_with_html(start_response, "<div>foo</div>")

    The default response status of "200 OK" can be overridden with the
    "status" keyword argument.
    """

    encoded = html.encode("utf-8")
    return respond_with_content(
        start_response,
        encoded,
        status=status,
        content_type="text/html; charset=utf-8",
        extra_headers=extra_headers,
    )


def created_at(
    request: Request,
    start_response: StartResponse,
    url_part: str,
    *,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """Prepare a 201 Created WSGI response with a Location header.

    The default content-type is "text/html" and the return value generates
    a simple HTML body.
    """

    url = absolute_url(request, url_part)
    html = created_at_page(url)
    all_headers = [_location_header(request, url_part)] + list(extra_headers)
    return respond_with_html(
        start_response,
        html,
        status=HTTPStatus.CREATED,
        extra_headers=all_headers,
    )


def created_as_json(
    request: Request,
    start_response: StartResponse,
    url_part: str,
    json: str | bytes | Any,
    *,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    """
    Prepare a 201 Created WSGI response with a Location header and JSON body.
    """

    all_headers = [_location_header(request, url_part)] + list(extra_headers)
    return respond_with_json(
        start_response,
        json,
        status=HTTPStatus.CREATED,
        extra_headers=all_headers,
    )


def temporary_redirect(
    request: Request,
    start_response: StartResponse,
    url_part: str,
    *,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    url = absolute_url(request, url_part)
    html = temporary_redirect_page(url)
    all_headers = [_location_header(request, url_part)] + list(extra_headers)
    return respond_with_html(
        start_response,
        html,
        status=HTTPStatus.TEMPORARY_REDIRECT,
        extra_headers=all_headers,
    )


def see_other(
    request: Request,
    start_response: StartResponse,
    url_part: str,
    *,
    extra_headers: Sequence[Header] = [],
) -> Iterable[bytes]:
    url = absolute_url(request, url_part)
    html = see_other_page(url)
    all_headers = [_location_header(request, url_part)] + list(extra_headers)
    return respond_with_html(
        start_response,
        html,
        status=HTTPStatus.SEE_OTHER,
        extra_headers=all_headers,
    )



from __future__ import annotations

import logging
from collections.abc import Iterable, Sequence
from http import HTTPStatus
from typing import Any

from asserts import (
    assert_equal,
    assert_in,
    assert_is_instance,
    assert_raises,
    assert_regex,
    fail,
)
from dectest import TestCase, before, test
from werkzeug.datastructures import WWWAuthenticate
from werkzeug.exceptions import Unauthorized
from werkzeug.wrappers import Request

from rouver.exceptions import ArgumentsError
from rouver.router import LOGGER_NAME, Router
from rouver.types import (
    StartResponse,
    WSGIApplication,
    WSGIEnvironment,
    WSGIResponse,
)
from rouver_test.testutil import StubStartResponse, default_environment


def handle_success(
    _: WSGIEnvironment, start_response: StartResponse
) -> Iterable[bytes]:
    start_response("200 OK", [])
    return [b""]


def handle_empty_path(
    environ: WSGIEnvironment, start_response: StartResponse
) -> Iterable[bytes]:
    assert_equal([], environ["rouver.path_args"])
    start_response("200 OK", [])
    return [b""]


def fail_if_called(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
    raise AssertionError("handler should not be called")


class RouterTest(TestCase):
    @before
    def setup_router(self) -> None:
        self.router = Router()
        self.router.error_handling = False
        self.start_response = StubStartResponse()
        self.environment = default_environment()
        self.disable_logger()

    def disable_logger(self) -> None:
        logging.getLogger(LOGGER_NAME).disabled = True

    def _create_path_checker(self, expected_path: str) -> WSGIApplication:
        def handle(
            environ: WSGIEnvironment, sr: StartResponse
        ) -> Sequence[bytes]:
            assert_equal(expected_path, environ["PATH_INFO"])
            sr("200 OK", [])
            return []

        return handle

    def handle_wsgi(
        self, method: str = "GET", path: str = "/", *, script_name: str = ""
    ) -> Iterable[bytes]:
        self.environment["REQUEST_METHOD"] = method
        self.environment["PATH_INFO"] = path
        self.environment["SCRIPT_NAME"] = script_name
        return self.router(self.environment, self.start_response)

    @test
    def not_found_response_page(self) -> None:
        response = self.handle_wsgi("GET", "/foo/bar")
        html = b"".join(response).decode("utf-8")
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>404 &#x2014; Not Found</title>
    </head>
    <body>
        <h1>404 &#x2014; Not Found</h1>
        <p>Path &#x27;/foo/bar&#x27; not found.</p>
    </body>
</html>
""",
            html,
        )

    @test
    def not_found_escape_path(self) -> None:
        response = self.handle_wsgi("GET", "/foo/<bar")
        html = b"".join(response).decode("utf-8")
        assert_in("<p>Path &#x27;/foo/&lt;bar&#x27; not found.</p>", html)

    @test
    def no_routes(self) -> None:
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8"
        )

    @test
    def handler_request(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal("test.example.com", environ["SERVER_NAME"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("", "GET", handle)])
        self.environment["SERVER_NAME"] = "test.example.com"
        self.handle_wsgi("GET", "")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def empty_route(self) -> None:
        self.router.add_routes([("", "GET", handle_empty_path)])
        self.handle_wsgi("GET", "")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def root_route(self) -> None:
        self.router.add_routes([("", "GET", handle_empty_path)])
        self.handle_wsgi("GET", "/")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def first_level(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal("", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return []

        self.router.add_routes([("foo", "GET", handle)])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def first_level__trailing_slash(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal("/", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return []

        self.router.add_routes([("foo", "GET", handle)])
        self.handle_wsgi("GET", "/foo/")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def first_level_wrong_path(self) -> None:
        self.router.add_routes([("foo", "GET", handle_empty_path)])
        self.handle_wsgi("GET", "/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def level_mismatch_1(self) -> None:
        self.router.add_routes([("foo/bar", "GET", handle_empty_path)])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def level_mismatch_2(self) -> None:
        self.router.add_routes([("foo", "GET", handle_empty_path)])
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def decode_path(self) -> None:
        self.router.add_routes([("foo/bär", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo/b%c3%a4r")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def invalid_path_encoding(self) -> None:
        self.router.add_routes([("foo/bar", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo/b%c3r")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    # Method Handling

    @test
    def wrong_method_response_page(self) -> None:
        self.router.add_routes(
            [("foo", "GET", fail_if_called), ("foo", "PUT", fail_if_called)]
        )
        response = self.handle_wsgi("POST", "/foo")
        html = b"".join(response).decode("utf-8")
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>405 &#x2014; Method Not Allowed</title>
    </head>
    <body>
        <h1>405 &#x2014; Method Not Allowed</h1>
        <p>Method &#x27;POST&#x27; not allowed. Please try GET or PUT.</p>
    </body>
</html>
""",
            html,
        )

    @test
    def wrong_method_escape_method(self) -> None:
        self.router.add_routes([("foo", "GET", fail_if_called)])
        response = self.handle_wsgi("G<T", "/foo")
        html = b"".join(response).decode("utf-8")
        assert_in(
            "<p>Method &#x27;G&lt;T&#x27; not allowed. Please try GET.</p>",
            html,
        )

    @test
    def wrong_method(self) -> None:
        self.router.add_routes(
            [("foo", "GET", fail_if_called), ("foo", "PUT", fail_if_called)]
        )
        self.handle_wsgi("POST", "/foo")
        self.start_response.assert_status(HTTPStatus.METHOD_NOT_ALLOWED)
        self.start_response.assert_header_equals("Allow", "GET, PUT")

    @test
    def wrong_method__multiple_matches(self) -> None:
        self.router.add_routes(
            [("foo", "GET", fail_if_called), ("foo", "GET", fail_if_called)]
        )
        self.handle_wsgi("POST", "/foo")
        self.start_response.assert_status(HTTPStatus.METHOD_NOT_ALLOWED)
        self.start_response.assert_header_equals("Allow", "GET")

    @test
    def call_right_method(self) -> None:
        self.router.add_routes(
            [
                ("foo", "GET", fail_if_called),
                ("foo", "POST", handle_success),
                ("foo", "PUT", fail_if_called),
            ]
        )
        self.handle_wsgi("POST", "/foo")
        self.start_response.assert_status(HTTPStatus.OK)

    # Path Templates

    @test
    def unknown_template(self) -> None:
        with assert_raises(KeyError):
            self.router.add_routes(
                [("foo/{unknown}/bar", "GET", fail_if_called)]
            )

    @test
    def no_template(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal([], environ["rouver.path_args"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("foo/bar", "GET", handle)])
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def template(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal(["xyzxyz"], environ["rouver.path_args"])
            start_response("200 OK", [])
            return [b""]

        def handle_path(
            request: Request, paths: Sequence[Any], path: str
        ) -> str:
            assert_is_instance(request, Request)
            assert_equal((), paths)
            return path * 2

        self.router.add_template_handler("handler", handle_path)

        self.router.add_routes([("foo/{handler}/bar", "GET", handle)])
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def multiple_templates(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal(["xyz", 123], environ["rouver.path_args"])
            start_response("200 OK", [])
            return [b""]

        def handle_path(_: Request, paths: Sequence[Any], __: str) -> int:
            assert_equal(("xyz",), paths)
            return 123

        self.router.add_template_handler("handler1", lambda _, __, ___: "xyz")
        self.router.add_template_handler("handler2", handle_path)

        self.router.add_routes(
            [("foo/{handler1}/bar/{handler2}", "GET", handle)]
        )
        self.handle_wsgi("GET", "/foo/xyz/bar/abc")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def template_handler_is_passed_decoded_value(self) -> None:
        def handle_path(_: Request, __: Any, v: str) -> None:
            assert_equal("foo/bar", v)

        self.router.add_template_handler("handler", handle_path)

        self.router.add_routes([("foo/{handler}", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo/foo%2Fbar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def template_handler_is_not_passed_an_invalid_value(self) -> None:
        def handle_path(_: Request, __: Any, v: str) -> None:
            fail("template handler should not have been called")

        self.router.add_template_handler("handler", handle_path)

        self.router.add_routes([("foo/{handler}", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo/foo%C3bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def template_value_error(self) -> None:
        def raise_value_error(_: Request, __: Sequence[str], ___: str) -> None:
            raise ValueError()

        self.router.add_template_handler("handler", raise_value_error)

        self.router.add_routes([("foo/{handler}/bar", "GET", fail_if_called)])
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def template_multiple_matches(self) -> None:
        def raise_value_error(_: Request, __: Sequence[str], ___: str) -> None:
            raise ValueError()

        self.router.add_template_handler("handler1", raise_value_error)
        self.router.add_template_handler("handler2", lambda _, __, ___: None)

        self.router.add_routes(
            [
                ("foo/{handler1}/bar", "GET", fail_if_called),
                ("foo/{handler2}/bar", "GET", handle_success),
            ]
        )
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def template_multiple_matches__match_first(self) -> None:
        self.router.add_template_handler("handler1", lambda _, __, ___: None)
        self.router.add_template_handler("handler2", lambda _, __, ___: None)

        self.router.add_routes(
            [
                ("foo/{handler1}/bar", "GET", handle_success),
                ("foo/{handler2}/bar", "GET", fail_if_called),
            ]
        )
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def template_call_once_per_value(self) -> None:
        calls = 0

        def increase_count(_: Request, __: Sequence[str], ___: str) -> None:
            nonlocal calls
            calls += 1

        self.router.add_template_handler("handler", increase_count)

        self.router.add_routes(
            [
                ("foo/{handler}/bar", "GET", fail_if_called),
                ("foo/{handler}/baz", "GET", handle_success),
            ]
        )
        self.handle_wsgi("GET", "/foo/xyz/baz")
        assert_equal(1, calls)

    @test
    def template_call_twice_for_differing_values(self) -> None:
        calls = 0

        def increase_count(_: Request, __: Sequence[str], ___: str) -> None:
            nonlocal calls
            calls += 1

        self.router.add_template_handler("handler", increase_count)

        self.router.add_routes(
            [
                ("foo/{handler}/bar", "GET", fail_if_called),
                ("foo/xyz/{handler}", "GET", handle_success),
            ]
        )
        self.handle_wsgi("GET", "/foo/xyz/baz")
        assert_equal(2, calls)

    # Wildcard Paths

    @test
    def no_wildcard_path(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal("", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("foo/bar", "GET", handle)])
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard_path__no_trailing_slash(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal([], environ["rouver.path_args"])
            assert_equal("", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("foo/bar/*", "GET", handle)])
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard_path__with_trailing_slash(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal([], environ["rouver.path_args"])
            assert_equal("/", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("foo/bar/*", "GET", handle)])
        self.handle_wsgi("GET", "/foo/bar/")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard_path__additional_path(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal([], environ["rouver.path_args"])
            assert_equal("/abc/def", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_routes([("foo/bar/*", "GET", handle)])
        self.handle_wsgi("GET", "/foo/bar/abc/def")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard_path__with_template(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal(["value"], environ["rouver.path_args"])
            assert_equal("/abc/def", environ["rouver.wildcard_path"])
            start_response("200 OK", [])
            return [b""]

        self.router.add_template_handler("bar", lambda *args: "value")
        self.router.add_routes([("foo/{bar}/*", "GET", handle)])
        self.handle_wsgi("GET", "/foo/unknown/abc/def")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard_path__too_short(self) -> None:
        self.router.add_routes([("foo/bar/*", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def wildcard_path__does_not_match(self) -> None:
        self.router.add_routes([("foo/bar/*", "GET", handle_success)])
        self.handle_wsgi("GET", "/foo/wrong")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def wildcard_path__not_at_end(self) -> None:
        with assert_raises(ValueError):
            self.router.add_routes([("foo/*/bar", "GET", handle_success)])

    @test
    def wildcard__before_more_specific(self) -> None:
        self.router.add_routes(
            [
                ("foo/*", "GET", handle_success),
                ("foo/bar", "GET", fail_if_called),
            ]
        )
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def wildcard__after_more_specific(self) -> None:
        self.router.add_routes(
            [
                ("foo/bar", "GET", handle_success),
                ("foo/*", "GET", fail_if_called),
            ]
        )
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    # Sub routers

    @test
    def sub_router(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("sub", "GET", self._create_path_checker("/sub"))])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar/sub")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__no_match(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("sub", "GET", fail_if_called)])
        self.router.add_sub_router("foo", sub)
        self.handle_wsgi("GET", "/wrong/sub")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def sub_router__base_with_slash(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("", "GET", self._create_path_checker("/"))])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar/")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__base_without_slash(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("", "GET", self._create_path_checker(""))])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__path_info(self) -> None:
        def app(env: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            assert_equal("/foo", env["PATH_INFO"])
            assert_equal("script/sub", env["SCRIPT_NAME"])
            assert_equal("/sub/foo", env["rouver.original_path_info"])
            sr("200 OK", [])
            return []

        self.router.error_handling = False
        self.router.add_sub_router("sub", app)
        self.handle_wsgi("GET", "/sub/foo", script_name="script")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__path_info_encoding(self) -> None:
        expected_path = "/föo".encode("utf-8").decode("latin-1")

        def app(env: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            assert_equal(expected_path, env["PATH_INFO"])
            sr("200 OK", [])
            return []

        self.router.error_handling = False
        self.router.add_sub_router("sub", app)
        self.handle_wsgi("GET", "/sub/föo".encode("utf-8").decode("latin-1"))
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__template_in_super_router(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal([], environ["rouver.path_args"])
            start_response("200 OK", [])
            return []

        def tmpl(_: Request, path: Sequence[str], v: str) -> str:
            assert_equal((), path)
            return v * 2

        sub = Router()
        sub.error_handling = False
        sub.add_template_handler("tmpl", tmpl)
        sub.add_routes([("sub", "GET", handle)])
        self.router.add_template_handler("tmpl", tmpl)
        self.router.add_sub_router("foo/{tmpl}", sub)
        self.handle_wsgi("GET", "/foo/bar/sub")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__template_in_sub_router(self) -> None:
        def handle(
            environ: WSGIEnvironment, start_response: StartResponse
        ) -> Iterable[bytes]:
            assert_equal(["xyzxyz"], environ["rouver.path_args"])
            start_response("200 OK", [])
            return []

        def tmpl(_: Request, path: Sequence[str], v: str) -> str:
            assert_equal((), path)
            return v * 2

        sub = Router()
        sub.error_handling = False
        sub.add_template_handler("tmpl", tmpl)
        sub.add_routes([("{tmpl}", "GET", handle)])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar/xyz")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__path_component(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("sub", "GET", handle_success)])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/barsub")
        self.start_response.assert_status(HTTPStatus.NOT_FOUND)

    @test
    def sub_router__match_after_other_routes(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("sub", "GET", fail_if_called)])
        self.router.add_routes([("foo/bar/sub", "GET", handle_success)])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar/sub")
        self.start_response.assert_status(HTTPStatus.OK)

    @test
    def sub_router__accepts_any_wsgi_app(self) -> None:
        def sub(environ: WSGIEnvironment, sr: StartResponse) -> WSGIResponse:
            assert_equal("/sub", environ["PATH_INFO"])
            sr("204 No Content", [])
            return []

        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/bar/sub")
        self.start_response.assert_status(HTTPStatus.NO_CONTENT)

    @test
    def sub_router__escaped_path(self) -> None:
        sub = Router()
        sub.error_handling = False
        sub.add_routes([("sub", "GET", self._create_path_checker("/s%75b"))])
        self.router.add_sub_router("foo/bar", sub)
        self.handle_wsgi("GET", "/foo/b%61r/s%75b")
        self.start_response.assert_status(HTTPStatus.OK)

    # Error Handling

    @test
    def internal_error_page(self) -> None:
        def handle(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            raise KeyError("Custom < error")

        self.router.error_handling = True
        self.router.add_routes([("foo", "GET", handle)])
        response = self.handle_wsgi("GET", "/foo")
        html = b"".join(response).decode("utf-8")
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>500 &#x2014; Internal Server Error</title>
    </head>
    <body>
        <h1>500 &#x2014; Internal Server Error</h1>
        <p>Internal server error.</p>
    </body>
</html>
""",
            html,
        )

    @test
    def template_key_error_with_error_handling(self) -> None:
        def raise_key_error(_: Request, __: Sequence[str], ___: str) -> None:
            raise KeyError()

        self.router.add_template_handler("handler", raise_key_error)

        self.router.add_routes([("foo/{handler}/bar", "GET", fail_if_called)])
        self.router.error_handling = True
        self.handle_wsgi("GET", "/foo/xyz/bar")
        self.start_response.assert_status(HTTPStatus.INTERNAL_SERVER_ERROR)

    @test
    def template_key_error_without_error_handling(self) -> None:
        def raise_key_error(_: Request, __: Sequence[str], ___: str) -> None:
            raise KeyError()

        self.router.add_template_handler("handler", raise_key_error)

        self.router.add_routes([("foo/{handler}/bar", "GET", fail_if_called)])
        self.router.error_handling = False
        with assert_raises(KeyError):
            self.handle_wsgi("GET", "/foo/xyz/bar")

    @test
    def handler_key_error_with_error_handling(self) -> None:
        def handle(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            raise KeyError()

        self.router.error_handling = True
        self.router.add_routes([("foo", "GET", handle)])
        self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.INTERNAL_SERVER_ERROR)

    @test
    def handler_key_error_without_error_handling(self) -> None:
        def handle(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            raise KeyError()

        self.router.add_routes([("foo", "GET", handle)])
        self.router.error_handling = False
        with assert_raises(KeyError):
            self.handle_wsgi("GET", "/foo")

    @test
    def http_error(self) -> None:
        def handle(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            raise Unauthorized(
                "Foo < Bar", www_authenticate=WWWAuthenticate("Test")
            )

        self.router.error_handling = False
        self.router.add_routes([("foo", "GET", handle)])
        response = self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.UNAUTHORIZED)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8"
        )
        self.start_response.assert_header_equals("WWW-Authenticate", "Test ")
        html = b"".join(response).decode("utf-8")
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>401 &#x2014; Unauthorized</title>
    </head>
    <body>
        <h1>401 &#x2014; Unauthorized</h1>
        <p>Foo &lt; Bar</p>
    </body>
</html>
""",
            html,
        )

    @test
    def arguments_error(self) -> None:
        def handle(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            raise ArgumentsError({"foo": "bar"})

        self.router.add_routes([("foo", "GET", handle)])
        response = self.handle_wsgi("GET", "/foo")
        self.start_response.assert_status(HTTPStatus.BAD_REQUEST)
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")
        assert_regex(
            html,
            r'<li class="argument">\s*'
            r'<span class="argument-name">foo</span>:\s*'
            r'<span class="error-message">bar</span>\s*'
            r"</li>",
        )



from http import HTTPStatus

from asserts import assert_equal
from dectest import TestCase, test

from rouver.status import status_line


class StatusLineTest(TestCase):
    @test
    def status_line(self) -> None:
        sl = status_line(HTTPStatus.NO_CONTENT)
        assert_equal("204 No Content", sl)



from __future__ import annotations

from collections.abc import Sequence
from enum import Enum
from io import BytesIO
from typing import IO, Any, Callable, Dict, Tuple, Union
from urllib.parse import parse_qs

from typing_extensions import Literal
from werkzeug.datastructures import FileStorage, MultiDict
from werkzeug.formparser import parse_form_data

from rouver.exceptions import ArgumentsError
from rouver.types import WSGIEnvironment


class Multiplicity(Enum):
    ANY = "0+"
    REQUIRED_ANY = "1+"
    REQUIRED = "1"
    OPTIONAL = "0-1"


ArgumentValueParser = Callable[[str], Any]
ArgumentValueType = Union[ArgumentValueParser, Literal["file", "file-or-str"]]
ArgumentTemplate = Tuple[str, ArgumentValueType, Multiplicity]
ArgumentDict = Dict[str, Any]

_GET_METHODS = ["GET", "HEAD"]
_FORM_METHODS = ["POST", "PUT", "PATCH", "DELETE"]


class _ArgumentError(Exception):
    pass


class FileArgument:
    """File argument result.

    This is a file-like object, containing a byte stream. It has additional
    fields "filename" and "content_type".

    >>> args = parse_args(..., [
    ...     ("file-arg", "file", Multiplicity.REQUIRED),
    ... ])
    >>> args["file-arg"].filename
    'my-file.txt'
    >>> args["file-arg"].content_type
    'text/plain'
    >>> args["file-arg"].read()
    b'file content'
    """

    def __init__(
        self, stream: IO[bytes], filename: str, content_type: str
    ) -> None:
        self._stream = stream
        self.filename = filename
        self.content_type = content_type

    def __getattr__(self, name: str) -> Any:
        return getattr(self._stream, name)


class ArgumentParser:
    """Parse CGI/WSGI arguments.

    As opposed to parse_args(), ArgumentParser.parse_args() can be called
    multiple times on the same instance during an request.
    """

    def __init__(self, environ: WSGIEnvironment) -> None:
        self.environ = environ

        method: str = environ.get("REQUEST_METHOD", "GET")
        args: MultiDict[str, Any]
        files: MultiDict[str, Any]
        if method in _GET_METHODS:
            qs: str = environ.get("QUERY_STRING", "")
            args = MultiDict(parse_qs(qs, keep_blank_values=True))
            files = MultiDict()
        elif method in _FORM_METHODS:
            _, args, files = parse_form_data(environ)
        else:
            raise ValueError("unsupported method: '{}'".format(method))
        self._arguments = _create_arg_dict(args, files)
        self._not_found = {a for a in self._arguments}

    def parse_args(
        self,
        argument_template: Sequence[ArgumentTemplate],
        *,
        exhaustive: bool = False,
    ) -> ArgumentDict:
        """Parse CGI/WSGI arguments and return an argument dict.

        argument_template is a list of (argname, value_parser, multiplicity)
        tuples. argname is the argument name that is expected to be provided.

        If multiplicity is REQUIRED and the argument was not supplied, an
        ArgumentsError is raised. If multiplicity is OPTIONAL and the
        argument was not supplied, the returned dict does not contain a key
        for this argument. multiplicity can also be ANY, in which case
        the key may be supplied any number of times, or REQUIRED_ANY, which
        works like ANY, but requires the argument to be provided at least
        once.

        value_parser is a callable that takes a single string argument. The
        value returned by this function will be used as the value of the
        returned dict. If the supplied string or list can't be parsed, this
        function should raise a ValueError exception. In this case, parse_args
        raises an ArgumentsError. If value_func is the string "file",
        the value of the argument is a FileArgument instance. If value_func
        is "file-or-str", the value is a FileArgument instance or a str
        if the argument was given, but without a file as argument.

        The returned dict contains the names of all parsed arguments that were
        present in the CGI arguments as key. The value is the corresponding
        parsed argument value. In the case of ANY and REQUIRED_ANY arguments,
        the value is a list of parsed arguments values. Each ANY argument
        has an entry in the dict, even if the argument was not supplied.
        In this case the value is the empty list.

        Arguments of HEAD and GET requests will be parsed from the query
        string, while arguments of POST, PUT, PATCH, and DELETE requests
        are parsed from the request body. Other methods are not supported.

        If the exhaustive argument is set to True and the request contains
        arguments not listed in the argument_template of this or previous
        calls, raise an ArgumentsError.

        If there is any error while parsing the arguments, an ArgumentsError
        will be raised.

        parse_args() consumes the request input, but the arguments are
        cached in the ArgumentParser object, so multiple calls on the same
        object are possible.
        """

        errors: dict[str, str] = {}
        parsed_arguments: ArgumentDict = {}

        def parse_template(
            name: str,
            value_parser: ArgumentValueType,
            multiplicity: Multiplicity,
        ) -> None:
            cls = _VALUE_PARSER_CLASSES[multiplicity]
            argument_parser = cls(self._arguments, name, value_parser)
            if argument_parser.should_parse():
                try:
                    parsed_arguments[name] = argument_parser.parse()
                except _ArgumentError as exc:
                    errors[name] = exc.args[0]
            self._not_found.discard(name)

        for tmpl in argument_template:
            parse_template(*tmpl)

        if exhaustive:
            for arg in self._not_found:
                errors[arg] = "unknown argument"

        if len(errors) > 0:
            raise ArgumentsError(errors)
        return parsed_arguments


class _Argument:
    def __init__(
        self, value: list[str] | FileStorage, *, is_file: bool = False
    ) -> None:
        self._value = value
        self.is_file = is_file

    def as_string(self) -> str:
        if not isinstance(self._value, list) or len(self._value) == 0:
            raise TypeError("value is not a string")
        return self._value[0]

    def as_list(self) -> list[str]:
        if not isinstance(self._value, list):
            raise TypeError("value is not a list of strings")
        return self._value

    def as_file(self) -> tuple[IO[bytes], str, str]:
        if not isinstance(self._value, FileStorage):
            raise TypeError("value is not a file")
        content_type = self._value.mimetype or "application/octet-stream"
        return self._value.stream, self._value.filename or "", content_type


def _create_arg_dict(
    args: MultiDict, files: MultiDict
) -> dict[str, _Argument]:
    _all_args: dict[str, _Argument] = {}
    for name, v in args.lists():
        _all_args[name] = _Argument(v)
    for name, v in files.items():
        _all_args[name] = _Argument(v, is_file=True)
    return _all_args


def parse_args(
    environ: WSGIEnvironment,
    argument_template: Sequence[ArgumentTemplate],
    *,
    exhaustive: bool = False,
) -> ArgumentDict:
    """Parse CGI/WSGI arguments and return an argument dict.

    argument_template is a list of (argname, value_parser, multiplicity)
    tuples. argname is the argument name that is expected to be provided.

    If multiplicity is REQUIRED and the argument was not supplied, an
    ArgumentsError is raised. If multiplicity is OPTIONAL and the
    argument was not supplied, the returned dict does not contain a key
    for this argument. multiplicity can also be ANY, in which case
    the key may be supplied any number of times, or REQUIRED_ANY, which works
    like ANY, but requires the argument to be provided at least once.

    value_parser is a callable that takes a single string argument. The value
    returned by this function will be used as the value of the returned
    dict. If the supplied string or list can't be parsed, this function
    should raise a ValueError exception. In this case, parse_args raises
    an ArgumentsError. If value_func is the string "file", the value of the
    argument is a FileArgument instance. If value_func is "file-or-str",
    the value is a FileArgument instance or a str if the argument was given,
    but without a file as argument.

    The returned dict contains the names of all parsed arguments that were
    present in the CGI arguments as key. The value is the corresponding
    parsed argument value. In the case of ANY and REQUIRED_ANY arguments,
    the value is a list of parsed arguments values. Each ANY argument has an
    entry in the dict, even if the argument was not supplied. In this case the
    value is the empty list.

    Arguments of HEAD and GET requests will be parsed from the query string,
    while arguments of POST, PUT, PATCH, and DELETE requests are parsed from
    the request body. Other methods are not supported.

    If the exhaustive argument is set to True and the request contains
    arguments not listed in the argument_template, raise an ArgumentsError.

    If there is any error while parsing the arguments, an ArgumentsError
    will be raised.

    >>> from io import StringIO
    >>> environment = {
    ...     "QUERY_STRING": "key=value&multi=1&multi=2",
    ...     "wsgi.input": StringIO(""),
    ... }
    >>> args = parse_args(environment, [
    ...     ("key", str, Multiplicity.REQUIRED),
    ...     ("multi", int, Multiplicity.ANY),
    ...     ("optional", str, Multiplicity.OPTIONAL),
    ... ])
    >>> args["key"]
    'value'
    >>> args["multi"]
    [1, 2]
    >>> "optional" in args
    False
    >>> parse_args(environment, [("key", int, Multiplicity.REQUIRED)])
    Traceback (most recent call last):
    ...
    ArgumentsError: 400 Bad Request
    >>> parse_args(environment, [("missing", str, Multiplicity.REQUIRED)])
    Traceback (most recent call last):
    ...
    ArgumentsError: 400 Bad Request
    >>> parse_args(
    ...     environment, [("multi", str, Multiplicity.ANY)], exhaustive=True
    ... )
    Traceback (most recent call last):
    ...
    ArgumentsError: 400 Bad Request

    parse_args() consumes the request input, so multiple calls per request
    are not possible.
    """

    parser = ArgumentParser(environ)
    return parser.parse_args(argument_template, exhaustive=exhaustive)


class _ValueParserWrapper:
    def parse_from_string(self, s: str) -> Any:
        raise NotImplementedError()

    def parse_from_file(
        self, stream: IO[bytes], filename: str, content_type: str
    ) -> Any:
        raise NotImplementedError()


class _FunctionValueParser(_ValueParserWrapper):
    def __init__(self, value_parser: ArgumentValueParser) -> None:
        super().__init__()
        self.value_parser = value_parser

    def parse_from_string(self, value: str) -> Any:
        try:
            return self.value_parser(value)
        except ValueError as exc:
            raise _ArgumentError(str(exc)) from exc

    def parse_from_file(
        self, stream: IO[bytes], filename: str, content_type: str
    ) -> Any:
        return self.parse_from_string(stream.read().decode("utf-8"))


class _FileArgumentValueParser(_ValueParserWrapper):
    def parse_from_string(self, value: str) -> FileArgument:
        stream = BytesIO(value.encode("utf-8"))
        return FileArgument(stream, "", "application/octet-stream")

    def parse_from_file(
        self, stream: IO[bytes], filename: str, content_type: str
    ) -> FileArgument:
        return FileArgument(stream, filename, content_type)


class _OptionalFileArgumentValueParser(_ValueParserWrapper):
    def parse_from_string(self, value: str) -> str:
        return value

    def parse_from_file(
        self, stream: IO[bytes], filename: str, content_type: str
    ) -> FileArgument:
        return FileArgument(stream, filename, content_type)


def _create_argument_value_parser(
    value_parser: ArgumentValueType,
) -> _ValueParserWrapper:
    if value_parser == "file":
        return _FileArgumentValueParser()
    elif value_parser == "file-or-str":
        return _OptionalFileArgumentValueParser()
    elif callable(value_parser):
        return _FunctionValueParser(value_parser)
    else:
        raise TypeError("invalid value parser '{}'".format(value_parser))


class _ArgumentParser:
    def __init__(
        self,
        args: dict[str, _Argument],
        name: str,
        value_parser: ArgumentValueType,
    ) -> None:
        self.args = args
        self.name = name
        self.value_parser = _create_argument_value_parser(value_parser)

    def should_parse(self) -> bool:
        return True

    def parse(self) -> Any:
        raise NotImplementedError()


class _SingleArgumentParser(_ArgumentParser):
    def parse(self) -> Any:
        raise NotImplementedError()

    def parse_single_arg(self) -> Any:
        arg = self.args[self.name]
        if arg.is_file:
            stream, filename, content_type = arg.as_file()
            return self.value_parser.parse_from_file(
                stream, filename, content_type
            )
        else:
            value = arg.as_string()
            return self.value_parser.parse_from_string(value)

    @property
    def arg_supplied(self) -> bool:
        try:
            return self.name in self.args
        except TypeError:
            return False


class _RequiredArgumentParser(_SingleArgumentParser):
    def parse(self) -> Any:
        if not self.arg_supplied:
            raise _ArgumentError("mandatory argument missing")
        return self.parse_single_arg()


class _OptionalArgumentParser(_SingleArgumentParser):
    def should_parse(self) -> bool:
        return self.arg_supplied

    def parse(self) -> Any:
        assert self.should_parse()
        return self.parse_single_arg()


class _MultiArgumentParser(_ArgumentParser):
    def parse(self) -> list[Any]:
        try:
            arg = self.args[self.name]
            values = arg.as_list()
        except KeyError:
            values = []
        return [self.value_parser.parse_from_string(v) for v in values]


class _AtLeastOneArgumentParser(_MultiArgumentParser):
    def parse(self) -> list[Any]:
        values = super().parse()
        if not values:
            raise _ArgumentError("mandatory argument missing")
        return values


_VALUE_PARSER_CLASSES = {
    Multiplicity.ANY: _MultiArgumentParser,
    Multiplicity.REQUIRED_ANY: _AtLeastOneArgumentParser,
    Multiplicity.REQUIRED: _RequiredArgumentParser,
    Multiplicity.OPTIONAL: _OptionalArgumentParser,
}



from http import HTTPStatus
from json import loads as json_decode

from asserts import assert_equal, assert_in, assert_raises
from dectest import TestCase, before, test
from werkzeug.wrappers import Request

from rouver.response import (
    created_as_json,
    created_at,
    respond,
    respond_with_content,
    respond_with_html,
    respond_with_json,
    see_other,
    temporary_redirect,
)
from rouver_test.testutil import StubStartResponse, default_environment


class RespondTest(TestCase):
    @test
    def default_status(self) -> None:
        sr = StubStartResponse()
        respond(sr)
        sr.assert_status(HTTPStatus.OK)

    @test
    def custom_status(self) -> None:
        sr = StubStartResponse()
        respond(sr, status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    @test
    def extra_headers(self) -> None:
        sr = StubStartResponse()
        respond(sr, extra_headers=[("X-Custom-Header", "Foobar")])
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    @test
    def no_content_type(self) -> None:
        sr = StubStartResponse()
        respond(sr)
        sr.assert_header_missing("Content-Type")

    @test
    def content_type(self) -> None:
        sr = StubStartResponse()
        respond(sr, content_type="image/png")
        sr.assert_header_equals("Content-Type", "image/png")

    @test
    def content_type_in_extra_headers(self) -> None:
        sr = StubStartResponse()
        respond(sr, extra_headers=[("Content-Type", "image/png")])
        sr.assert_header_equals("Content-Type", "image/png")

    @test
    def error_if_content_type_also_in_extra_headers(self) -> None:
        sr = StubStartResponse()
        with assert_raises(ValueError):
            respond(
                sr,
                content_type="image/png",
                extra_headers=[("Content-Type", "image/jpeg")],
            )

    @test
    def response(self) -> None:
        sr = StubStartResponse()
        response = respond(sr)
        assert_equal(b"", b"".join(response))


class RespondWithContentTest(TestCase):
    @test
    def default_status(self) -> None:
        sr = StubStartResponse()
        respond_with_content(sr, b"")
        sr.assert_status(HTTPStatus.OK)

    @test
    def custom_status(self) -> None:
        sr = StubStartResponse()
        respond_with_content(sr, b"", status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    @test
    def default_content_type(self) -> None:
        sr = StubStartResponse()
        respond_with_content(sr, b"")
        sr.assert_header_equals("Content-Type", "application/octet-stream")

    @test
    def custom_content_type(self) -> None:
        sr = StubStartResponse()
        respond_with_content(sr, b"", content_type="text/plain")
        sr.assert_header_equals("Content-Type", "text/plain")

    @test
    def content_length(self) -> None:
        sr = StubStartResponse()
        respond_with_content(sr, b"foobar")
        sr.assert_header_equals("Content-Length", "6")

    @test
    def extra_headers(self) -> None:
        sr = StubStartResponse()
        respond_with_content(
            sr, b"", extra_headers=[("X-Custom-Header", "Foobar")]
        )
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    @test
    def return_value(self) -> None:
        sr = StubStartResponse()
        response = respond_with_content(sr, b"foobar")
        assert_equal(b"foobar", b"".join(response))


class RespondWithJSONTest(TestCase):
    @test
    def default_status(self) -> None:
        sr = StubStartResponse()
        respond_with_json(sr, {})
        sr.assert_status(HTTPStatus.OK)

    @test
    def custom_status(self) -> None:
        sr = StubStartResponse()
        respond_with_json(sr, {}, status=HTTPStatus.NOT_ACCEPTABLE)
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    @test
    def content_type(self) -> None:
        sr = StubStartResponse()
        respond_with_json(sr, {})
        sr.assert_header_equals(
            "Content-Type", "application/json; charset=utf-8"
        )

    @test
    def content_length(self) -> None:
        sr = StubStartResponse()
        respond_with_json(sr, {"foo": 33})
        sr.assert_header_equals("Content-Length", "11")

    @test
    def extra_headers(self) -> None:
        sr = StubStartResponse()
        respond_with_json(
            sr, {}, extra_headers=[("X-Custom-Header", "Foobar")]
        )
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    @test
    def json_as_bytes(self) -> None:
        sr = StubStartResponse()
        response = respond_with_json(sr, b'{"foo": 3}')
        assert_equal(b'{"foo": 3}', b"".join(response))

    @test
    def json_as_str(self) -> None:
        sr = StubStartResponse()
        response = respond_with_json(sr, '{"föo": 3}')
        assert_equal('{"föo": 3}'.encode("utf-8"), b"".join(response))

    @test
    def json_as_object(self) -> None:
        sr = StubStartResponse()
        response = respond_with_json(sr, {"föo": 3})
        assert_equal(b'{"f\\u00f6o": 3}', b"".join(response))


class RespondWithHTMLTest(TestCase):
    @test
    def default_status(self) -> None:
        sr = StubStartResponse()
        respond_with_html(sr, "<div>Test</div>")
        sr.assert_status(HTTPStatus.OK)

    @test
    def custom_status(self) -> None:
        sr = StubStartResponse()
        respond_with_html(
            sr, "<div>Test</div>", status=HTTPStatus.NOT_ACCEPTABLE
        )
        sr.assert_status(HTTPStatus.NOT_ACCEPTABLE)

    @test
    def content_type(self) -> None:
        sr = StubStartResponse()
        respond_with_html(sr, "<div>Test</div>")
        sr.assert_header_equals("Content-Type", "text/html; charset=utf-8")

    @test
    def content_length(self) -> None:
        sr = StubStartResponse()
        respond_with_html(sr, "<div>Test</div>")
        sr.assert_header_equals("Content-Length", "15")

    @test
    def extra_headers(self) -> None:
        sr = StubStartResponse()
        respond_with_html(
            sr,
            "<div>Täst</div>",
            extra_headers=[("X-Custom-Header", "Foobar")],
        )
        sr.assert_header_equals("X-Custom-Header", "Foobar")

    @test
    def return_value(self) -> None:
        sr = StubStartResponse()
        response = respond_with_html(sr, "<div>Test</div>")
        assert_equal(b"<div>Test</div>", b"".join(response))

    @test
    def return_value_encoding(self) -> None:
        sr = StubStartResponse()
        response = respond_with_html(sr, "<div>Täst</div>")
        assert_equal("<div>Täst</div>".encode("utf-8"), b"".join(response))


class CreatedAtTest(TestCase):
    @before
    def setup_environment(self) -> None:
        self.environment = default_environment()
        self.start_response = StubStartResponse()

    @test
    def headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_at(request, self.start_response, "/foo/bar")
        self.start_response.assert_status(HTTPStatus.CREATED)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8"
        )
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def absolute_url(self) -> None:
        request = Request(self.environment)
        created_at(request, self.start_response, "http://example.com/foo")
        self.start_response.assert_header_equals(
            "Location", "http://example.com/foo"
        )

    @test
    def url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_at(request, self.start_response, "foo/bar")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def umlauts_in_url(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_at(request, self.start_response, "foo/bär")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/b%C3%A4r"
        )

    @test
    def extra_headers(self) -> None:
        request = Request(self.environment)
        created_at(
            request,
            self.start_response,
            "foo",
            extra_headers=[("X-Foo", "Bar")],
        )
        self.start_response.assert_header_equals("X-Foo", "Bar")

    @test
    def html(self) -> None:
        request = Request(self.environment)
        response = created_at(request, self.start_response, "foo/bar")
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")


class CreatedAsJSONTest(TestCase):
    @before
    def setup_environment(self) -> None:
        self.environment = default_environment()
        self.start_response = StubStartResponse()

    @test
    def headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_as_json(request, self.start_response, "/foo/bar", {})
        self.start_response.assert_status(HTTPStatus.CREATED)
        self.start_response.assert_header_equals(
            "Content-Type", "application/json; charset=utf-8"
        )
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def absolute_url(self) -> None:
        request = Request(self.environment)
        created_as_json(
            request, self.start_response, "http://example.com/foo", {}
        )
        self.start_response.assert_header_equals(
            "Location", "http://example.com/foo"
        )

    @test
    def url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_as_json(request, self.start_response, "foo/bar", {})
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def umlauts_in_url(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        created_as_json(request, self.start_response, "foo/bär", {})
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/b%C3%A4r"
        )

    @test
    def extra_headers(self) -> None:
        request = Request(self.environment)
        created_as_json(
            request,
            self.start_response,
            "foo",
            {},
            extra_headers=[("X-Foo", "Bar")],
        )
        self.start_response.assert_header_equals("X-Foo", "Bar")

    @test
    def json(self) -> None:
        request = Request(self.environment)
        response = created_as_json(
            request, self.start_response, "foo/bar", {"foo": 3}
        )
        json = json_decode(b"".join(response).decode("utf-8"))
        assert_equal({"foo": 3}, json)


class TemporaryRedirectTest(TestCase):
    @before
    def setup_environment(self) -> None:
        self.environment = default_environment()
        self.start_response = StubStartResponse()

    @test
    def headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        temporary_redirect(request, self.start_response, "/foo/bar")
        self.start_response.assert_status(HTTPStatus.TEMPORARY_REDIRECT)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8"
        )
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def absolute_url(self) -> None:
        request = Request(self.environment)
        temporary_redirect(
            request, self.start_response, "http://example.com/foo"
        )
        self.start_response.assert_header_equals(
            "Location", "http://example.com/foo"
        )

    @test
    def url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        temporary_redirect(request, self.start_response, "foo/bar")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def umlauts_in_url(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        temporary_redirect(request, self.start_response, "foo/bär")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/b%C3%A4r"
        )

    @test
    def do_not_encode_cgi_arguments(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        temporary_redirect(
            request, self.start_response, "foo?bar=baz&abc=%6A;+,@:$"
        )
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo?bar=baz&abc=%6A;+,@:$"
        )

    @test
    def extra_headers(self) -> None:
        request = Request(self.environment)
        temporary_redirect(
            request,
            self.start_response,
            "foo",
            extra_headers=[("X-Foo", "Bar")],
        )
        self.start_response.assert_header_equals("X-Foo", "Bar")

    @test
    def html(self) -> None:
        request = Request(self.environment)
        response = temporary_redirect(request, self.start_response, "foo/bar")
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")
        assert_in("http://www.example.com/foo/bar", html)


class SeeOtherTest(TestCase):
    @before
    def setup_environment(self) -> None:
        self.environment = default_environment()
        self.start_response = StubStartResponse()

    @test
    def headers(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        see_other(request, self.start_response, "/foo/bar")
        self.start_response.assert_status(HTTPStatus.SEE_OTHER)
        self.start_response.assert_header_equals(
            "Content-Type", "text/html; charset=utf-8"
        )
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/bar"
        )

    @test
    def absolute_url(self) -> None:
        request = Request(self.environment)
        see_other(request, self.start_response, "http://example.com/foo")
        self.start_response.assert_header_equals(
            "Location", "http://example.com/foo"
        )

    @test
    def url_without_leading_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        self.environment["PATH_INFO"] = "/abc/def/"
        request = Request(self.environment)
        see_other(request, self.start_response, "foo/bar")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/abc/def/foo/bar"
        )

    @test
    def url_path_without_trailing_slash(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        self.environment["PATH_INFO"] = "/abc/def"
        request = Request(self.environment)
        see_other(request, self.start_response, "foo/bar")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/abc/foo/bar"
        )

    @test
    def umlauts_in_url(self) -> None:
        self.environment["SERVER_NAME"] = "www.example.com"
        request = Request(self.environment)
        see_other(request, self.start_response, "foo/bär")
        self.start_response.assert_header_equals(
            "Location", "http://www.example.com/foo/b%C3%A4r"
        )

    @test
    def extra_headers(self) -> None:
        request = Request(self.environment)
        see_other(
            request,
            self.start_response,
            "foo",
            extra_headers=[("X-Foo", "Bar")],
        )
        self.start_response.assert_header_equals("X-Foo", "Bar")

    @test
    def html(self) -> None:
        request = Request(self.environment)
        response = see_other(request, self.start_response, "foo/bar")
        html = b"".join(response).decode("utf-8")
        assert html.startswith("<!DOCTYPE html>")
        assert_in("http://www.example.com/foo/bar", html)



from http import HTTPStatus

from asserts import assert_equal, assert_in, assert_raises
from dectest import TestCase, test

from rouver.html import (
    bad_arguments_list,
    created_at_page,
    http_status_page,
    see_other_page,
    temporary_redirect_page,
)


class HTTPStatusPageTest(TestCase):
    @test
    def default(self) -> None:
        html = http_status_page(HTTPStatus.NOT_ACCEPTABLE)
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>406 &#x2014; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &#x2014; Not Acceptable</h1>
    </body>
</html>
""",
            html,
        )

    @test
    def message_and_html_message(self) -> None:
        with assert_raises(ValueError):
            http_status_page(
                HTTPStatus.NOT_ACCEPTABLE,
                message="Test",
                html_message="HTML Test",
            )

    @test
    def message(self) -> None:
        html = http_status_page(
            HTTPStatus.NOT_ACCEPTABLE, message="Test <em>message</em>."
        )
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>406 &#x2014; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &#x2014; Not Acceptable</h1>
        <p>Test &lt;em&gt;message&lt;/em&gt;.</p>
    </body>
</html>
""",
            html,
        )

    @test
    def html_message(self) -> None:
        html = http_status_page(
            HTTPStatus.NOT_ACCEPTABLE, html_message="Test <em>message</em>."
        )
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>406 &#x2014; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &#x2014; Not Acceptable</h1>
        <p>Test <em>message</em>.</p>
    </body>
</html>
""",
            html,
        )

    @test
    def html_content(self) -> None:
        html = http_status_page(
            HTTPStatus.NOT_ACCEPTABLE, html_content="<div>Test content.</div>"
        )
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>406 &#x2014; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &#x2014; Not Acceptable</h1>
<div>Test content.</div>
    </body>
</html>
""",
            html,
        )

    @test
    def html_message_and_html_content(self) -> None:
        html = http_status_page(
            HTTPStatus.NOT_ACCEPTABLE,
            html_message="Test message.",
            html_content="<div>Test content.</div>",
        )
        assert_equal(
            """<!DOCTYPE html>
<html>
    <head>
        <title>406 &#x2014; Not Acceptable</title>
    </head>
    <body>
        <h1>406 &#x2014; Not Acceptable</h1>
        <p>Test message.</p>
<div>Test content.</div>
    </body>
</html>
""",
            html,
        )


class CreatedAtPageTest(TestCase):
    @test
    def encode_url(self) -> None:
        page = created_at_page('/foo/"bar"')
        assert_in(
            '<a href="/foo/&quot;bar&quot;">/foo/&quot;bar&quot;</a>', page
        )


class TemporaryRedirectPageTest(TestCase):
    @test
    def encode_url(self) -> None:
        page = temporary_redirect_page('/foo/"bar"')
        assert_in(
            '<a href="/foo/&quot;bar&quot;">/foo/&quot;bar&quot;</a>', page
        )


class SeeOtherPageTest(TestCase):
    @test
    def encode_url(self) -> None:
        page = see_other_page('/foo/"bar"')
        assert_in(
            '<a href="/foo/&quot;bar&quot;">/foo/&quot;bar&quot;</a>', page
        )


class BadArgumentsListTest(TestCase):
    @test
    def empty_dict(self) -> None:
        html = bad_arguments_list({})
        assert_equal("", html)

    @test
    def one_item(self) -> None:
        html = bad_arguments_list({"foo": "bar"})
        assert_equal(
            """<ul class="bad-arguments">
    <li class="argument">
        <span class="argument-name">foo</span>:
        <span class="error-message">bar</span>
    </li>
</ul>
""",
            html,
        )

    @test
    def multiple_items_alphabetically(self) -> None:
        html = bad_arguments_list(
            {"def": "error 1", "abc": "error 2", "ghi": "error 3"}
        )
        assert_equal(
            """<ul class="bad-arguments">
    <li class="argument">
        <span class="argument-name">abc</span>:
        <span class="error-message">error 2</span>
    </li>
    <li class="argument">
        <span class="argument-name">def</span>:
        <span class="error-message">error 1</span>
    </li>
    <li class="argument">
        <span class="argument-name">ghi</span>:
        <span class="error-message">error 3</span>
    </li>
</ul>
""",
            html,
        )

    @test
    def escape_html(self) -> None:
        html = bad_arguments_list({"a<c": "d<f"})
        assert_equal(
            """<ul class="bad-arguments">
    <li class="argument">
        <span class="argument-name">a&lt;c</span>:
        <span class="error-message">d&lt;f</span>
    </li>
</ul>
""",
            html,
        )



import pytest
from werkzeug import Request

from rouver.util import absolute_url, rfc5987_encode


class TestAbsoluteURL:
    @staticmethod
    def _create_request(*, path_info: str = "/path") -> Request:
        return Request(
            {
                "wsgi.url_scheme": "https",
                "SERVER_NAME": "example.com",
                "SERVER_PORT": "443",
                "SCRIPT_NAME": "/base/",
                "PATH_INFO": path_info,
            }
        )

    def test_path_is_not_ascii(self) -> None:
        request = self._create_request()
        assert (
            absolute_url(request, "/~föo") == "https://example.com/~f%C3%B6o"
        )

    def test_path_is_absolute(self) -> None:
        request = self._create_request()
        assert (
            absolute_url(request, "https://example.org/foo")
            == "https://example.org/foo"
        )

    def test_path_is_root_relative(self) -> None:
        request = self._create_request()
        assert absolute_url(request, "/foo") == "https://example.com/foo"

    def test_path_is_relative__base_with_slash(self) -> None:
        request = self._create_request(path_info="/path/")
        assert (
            absolute_url(request, "foo") == "https://example.com/base/path/foo"
        )

    def test_path_is_relative__base_without_slash(self) -> None:
        request = self._create_request(path_info="/path")
        assert absolute_url(request, "foo") == "https://example.com/base/foo"

    def test_do_not_encode_special_characters(self) -> None:
        request = self._create_request()
        assert (
            absolute_url(request, "/foo?bar=baz&abc=%6A;+,@:$")
            == "https://example.com/foo?bar=baz&abc=%6A;+,@:$"
        )


@pytest.mark.parametrize(
    "string, encoded",
    [
        ("", "key="),
        ("foo", "key=foo"),
        ("foo bar\tbaz", 'key="foo bar\\\tbaz"'),
        ("föo bar", "key*=UTF-8''f%C3%B6o%20bar"),
    ],
)
def test_rfc5987_encode(string: str, encoded: str) -> None:
    assert encoded == rfc5987_encode("key", string)



from asserts import assert_equal
from dectest import TestCase, test

from rouver.exceptions import ArgumentsError


class ArgumentsErrorTest(TestCase):
    @test
    def message(self) -> None:
        error = ArgumentsError({"foo": "bar"})
        assert_equal("400 Bad Request: invalid arguments", str(error))

    @test
    def arguments(self) -> None:
        error = ArgumentsError({"foo": "bar"})
        assert_equal({"foo": "bar"}, error.arguments)



from __future__ import annotations

import sys
from collections.abc import Iterable, Sequence
from http import HTTPStatus
from io import BytesIO
from typing import Any

import pytest
from werkzeug.formparser import parse_form_data

from rouver.args import ArgumentTemplate, Multiplicity, parse_args
from rouver.exceptions import ArgumentsError
from rouver.test import (
    ArgumentToTest,
    FakeRequest,
    FakeResponse,
    create_request,
    run_wsgi_test,
    test_wsgi_arguments as run_wsgi_arguments_test,
)
from rouver.types import StartResponse, WSGIApplication, WSGIEnvironment


def assert_wsgi_input_stream(stream: object) -> None:
    assert hasattr(stream, "read")
    assert hasattr(stream, "readline")
    assert hasattr(stream, "readlines")
    assert hasattr(stream, "__iter__")


class TestRequestTest:
    def test_attributes(self) -> None:
        request = create_request("GET", "/foo/bar")
        assert request.method == "GET"
        assert request.path == "/foo/bar"
        assert request.content_type is None
        assert isinstance(request.error_stream, BytesIO)

    def test_capitalize_method(self) -> None:
        request = create_request("pOst", "/foo/bar")
        assert request.method == "POST"

    def test_to_environment__minimal(self) -> None:
        request = create_request("GET", "/foo/bar")
        environ = request.to_environment()
        assert {
            "REQUEST_METHOD": "GET",
            "PATH_INFO": "/foo/bar",
            "SERVER_NAME": "www.example.com",
            "SERVER_PORT": "80",
            "SERVER_PROTOCOL": "HTTP/1.1",
            "REMOTE_ADDR": "127.0.0.1",
            "SCRIPT_NAME": "",
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": "http",
            "wsgi.multithread": False,
            "wsgi.multiprocess": False,
            "wsgi.run_once": True,
            "wsgi.errors": request.error_stream,
        }.items() <= environ.items()
        assert_wsgi_input_stream(environ["wsgi.input"])
        assert environ["wsgi.input"].read() == b""
        assert "CONTENT_TYPE" not in environ
        assert "CONTENT_LENGTH" not in environ
        assert "QUERY_STRING" not in environ

    def test_to_environment__post(self) -> None:
        request = create_request("POST", "/foo/bar")
        environ = request.to_environment()
        assert {
            "REQUEST_METHOD": "POST",
            "PATH_INFO": "/foo/bar",
            "SCRIPT_NAME": "",
            "SERVER_NAME": "www.example.com",
            "SERVER_PORT": "80",
            "SERVER_PROTOCOL": "HTTP/1.1",
            "REMOTE_ADDR": "127.0.0.1",
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": "http",
            "wsgi.multithread": False,
            "wsgi.multiprocess": False,
            "wsgi.run_once": True,
            "wsgi.errors": request.error_stream,
        }.items() <= environ.items()
        assert_wsgi_input_stream(environ["wsgi.input"])
        assert environ["wsgi.input"].read() == b""
        assert "CONTENT_TYPE" not in environ
        assert "CONTENT_LENGTH" not in environ
        assert "QUERY_STRING" not in environ

    def test_to_environment__post_urlencoded(self) -> None:
        request = create_request("POST", "/foo/bar")
        request.add_argument("arg", "value")
        environ = request.to_environment()
        assert {
            "REQUEST_METHOD": "POST",
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
            "CONTENT_LENGTH": "9",
        }.items() <= environ.items()
        assert_wsgi_input_stream(environ["wsgi.input"])
        assert environ["wsgi.input"].read() == b"arg=value"
        assert "QUERY_STRING" not in environ

    def test__explicit_script_name(self) -> None:
        request = create_request("GET", "/baz/bar")
        assert request.full_path == "/baz/bar"
        request.script_name = "/foo"
        assert request.full_path == "/foo/baz/bar"
        assert {
            "SCRIPT_NAME": "/foo",
            "PATH_INFO": "/baz/bar",
        }.items() <= request.to_environment().items()

    def test_set_env_var(self) -> None:
        request = create_request("GET", "/foo/bar")
        request.set_env_var("foo.bar", "baz")
        environ = request.to_environment()
        assert environ["foo.bar"] == "baz"

    def test_set_env_var__priority(self) -> None:
        request = create_request("GET", "/foo/bar")
        request.set_env_var("SERVER_PORT", "8888")
        request.set_env_var("HTTP_X_FOO", "Set by env var")
        request.set_header("X-Foo", "Set by header")
        environ = request.to_environment()
        assert environ["SERVER_PORT"] == "8888"
        assert environ["HTTP_X_FOO"] == "Set by env var"

    def test_set_header(self) -> None:
        request = create_request("GET", "/foo/bar")
        request.set_header("X-Foobar", "Baz")
        environ = request.to_environment()
        assert environ["HTTP_X_FOOBAR"] == "Baz"

    def test_set_header__content_type(self) -> None:
        request = create_request("GET", "/foo/bar")
        request.set_header("Content-Type", "text/html")
        assert request.content_type == "text/html"
        environ = request.to_environment()
        assert environ["CONTENT_TYPE"] == "text/html"
        assert "HTTP_CONTENT_TYPE" not in environ

    def test_add_argument__content_type(self) -> None:
        request = create_request("POST", "/foo/bar")
        assert request.content_type is None
        request.add_argument("foo", "bar")
        assert request.content_type is None
        environ = request.to_environment()
        assert environ["CONTENT_TYPE"] == "application/x-www-form-urlencoded"

        request = create_request("POST", "/foo/bar")
        request.content_type = "image/png"
        request.add_argument("abc", "def")
        assert request.content_type == "image/png"
        environ = request.to_environment()
        assert environ["CONTENT_TYPE"] == "image/png"

        request = create_request("GET", "/foo/bar")
        assert request.content_type is None
        request.add_argument("foo", "bar")
        assert request.content_type is None
        environ = request.to_environment()
        assert "CONTENT_TYPE" not in environ

    def test_add_file_argument__content_type(self) -> None:
        request = create_request("POST", "/foo/bar")
        assert request.content_type is None
        request.add_file_argument("foo", b"", "text/plain")
        assert request.content_type is None
        environ = request.to_environment()
        content_type, boundary = environ["CONTENT_TYPE"].split(";")
        assert content_type == "multipart/form-data"

        request = create_request("POST", "/foo/bar")
        request.content_type = "image/png"
        request.add_file_argument("abc", b"", "text/plain")
        assert request.content_type == "image/png"
        environ = request.to_environment()
        assert environ["CONTENT_TYPE"] == "image/png"

    def test_add_argument__body_set(self) -> None:
        put_request = create_request("PUT", "/foo")
        put_request.body = b"Body"
        with pytest.raises(ValueError):
            put_request.add_argument("foo", "bar")

    def test_add_file_argument__body_set(self) -> None:
        put_request = create_request("PUT", "/foo")
        put_request.body = b"Body"
        with pytest.raises(ValueError):
            put_request.add_file_argument("foo", b"", "text/plain")

    def test_add_file_argument__unsupported_method(self) -> None:
        get_request = create_request("GET", "/foo")
        with pytest.raises(ValueError):
            get_request.add_file_argument("foo", b"", "text/plain")

    def test_to_environment__content_type(self) -> None:
        request = create_request("GET", "/foo/bar")
        request.content_type = "image/png"
        environ = request.to_environment()
        assert environ["CONTENT_TYPE"] == "image/png"

    def test_arguments__get_request(self) -> None:
        request = create_request("GET", "/foo")
        request.add_argument("foo", "bar")
        request.add_argument("abc", ["def", "ghi"])
        environ = request.to_environment()
        assert environ["QUERY_STRING"] == "foo=bar&abc=def&abc=ghi"

    def test_arguments__put_request(self) -> None:
        request = create_request("PUT", "/foo")
        request.add_argument("foo", "bar")
        request.add_argument("abc", ["def", "ghi"])
        environ = request.to_environment()
        assert "QUERY_STRING" not in environ
        assert environ["CONTENT_TYPE"] == "application/x-www-form-urlencoded"
        assert environ["wsgi.input"].read() == b"foo=bar&abc=def&abc=ghi"

    def test_arguments__quote(self) -> None:
        request = create_request("GET", "/foo")
        request.add_argument("föo", "bär")
        environ = request.to_environment()
        assert environ["QUERY_STRING"] == "f%C3%B6o=b%C3%A4r"

    def test_file_arguments(self) -> None:
        request = create_request("PUT", "/foo")
        request.add_argument("foo", "bar")
        request.add_file_argument("file1", b"content1", "text/plain")
        request.add_file_argument(
            "file2", b"content2", "image/png", filename="foobar"
        )
        environ = request.to_environment()
        assert "QUERY_STRING" not in environ
        content_type, boundary = environ["CONTENT_TYPE"].split(";")
        assert content_type == "multipart/form-data"
        _, args, files = parse_form_data(environ)
        assert len(args) == 1
        assert "bar" == args["foo"]
        assert len(files) == 2
        file1 = files["file1"]
        assert file1.mimetype == "text/plain"
        assert file1.filename == ""
        assert file1.stream.read() == b"content1"
        file2 = files["file2"]
        assert file2.mimetype == "image/png"
        assert file2.filename == "foobar"
        assert file2.stream.read() == b"content2"

    def test_file_arguments__umlauts(self) -> None:
        request = create_request("PUT", "/foo")
        request.add_argument('f"öo', "bär")
        request.add_file_argument(
            'f"öle', b"", "text/plain", filename="ä\"'bc"
        )
        environ = request.to_environment()
        assert "QUERY_STRING" not in environ
        content_type, _ = environ["CONTENT_TYPE"].split(";")
        assert content_type == "multipart/form-data"
        _, args, files = parse_form_data(environ)
        assert len(args) == 1
        assert "bär" == args['f"%C3%B6o']
        assert len(files) == 1
        file = files['f"%C3%B6le']
        assert file.mimetype == "text/plain"
        assert file.filename == "ä\"'bc"
        assert file.stream.read() == b""

    def test_clear_arguments(self) -> None:
        request = create_request("GET", "/foo")
        request.add_argument("foo", "bar")
        request.clear_arguments()
        environ = request.to_environment()
        assert "QUERY_STRING" not in environ

        request = create_request("POST", "/foo")
        request.add_argument("foo", "bar")
        request.clear_arguments()
        environ = request.to_environment()
        content = environ["wsgi.input"].read()
        assert content == b""

    def test_body(self) -> None:
        request = create_request("POST", "/")
        assert request.body == b""
        request.body = b"Test Body"
        assert request.body == b"Test Body"
        environ = request.to_environment()
        assert environ.get("CONTENT_LENGTH") == "9"
        assert environ["wsgi.input"].read() == b"Test Body"

    def test_set_body_in_get_request(self) -> None:
        request = create_request("GET", "/")
        request.body = b""
        with pytest.raises(ValueError):
            request.body = b"Test Body"

    def test_set_body_when_argument_is_set(self) -> None:
        request = create_request("POST", "/")
        request.add_argument("foo", "bar")
        with pytest.raises(ValueError):
            request.body = b""
        with pytest.raises(ValueError):
            request.body = b"Body"

    def _assert_json_request(
        self, request: FakeRequest, expected_body: bytes
    ) -> None:
        assert request.body == expected_body
        env = request.to_environment()
        assert env["CONTENT_LENGTH"] == str(len(expected_body))
        assert env["CONTENT_TYPE"] == "application/json; charset=utf-8"
        assert env["wsgi.input"].read() == expected_body

    def test_set_json_request__get_request(self) -> None:
        request = create_request("GET", "/")
        with pytest.raises(ValueError):
            request.set_json_request(b"{}")

    def test_set_json_request__bytes(self) -> None:
        request = create_request("POST", "/")
        request.set_json_request(b"{}")
        self._assert_json_request(request, b"{}")

    def test_set_json_request__str(self) -> None:
        request = create_request("POST", "/")
        request.set_json_request('{"foo": "bär"}')
        self._assert_json_request(request, '{"foo": "bär"}'.encode("utf-8"))

    def test_set_json_request__dict(self) -> None:
        request = create_request("POST", "/")
        request.set_json_request({"foo": "bär"})
        self._assert_json_request(request, b'{"foo": "b\\u00e4r"}')

    def test_set_json_request__list(self) -> None:
        request = create_request("POST", "/")
        request.set_json_request(["foo", "bär"])
        self._assert_json_request(request, b'["foo", "b\\u00e4r"]')


class TestFakeResponse:
    def test_attributes(self) -> None:
        response = FakeResponse("200 OK", [])
        assert response.status_line == "200 OK"
        assert response.status == HTTPStatus.OK
        assert response.body == b""

    def test_unknown_status(self) -> None:
        with pytest.raises(ValueError):
            FakeResponse("999 Unknown", [])

    def test_invalid_status_line(self) -> None:
        with pytest.raises(ValueError):
            FakeResponse("INVALID", [])

    def test_get_header_value(self) -> None:
        response = FakeResponse(
            "200 OK",
            [
                ("X-Header", "Foobar"),
                ("Content-Type", "image/png"),
                ("Allow", "GET"),
            ],
        )
        assert response.get_header_value("Content-Type") == "image/png"
        assert response.get_header_value("content-TYPE") == "image/png"
        with pytest.raises(ValueError):
            response.get_header_value("X-Unknown")

    def test_parse_json_body(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "application/json")]
        )
        response.body = b'{"foo": 5}'
        json = response.parse_json_body()
        assert json == {"foo": 5}

    def test_parse_json_body__wrong_content_type(self) -> None:
        response = FakeResponse("200 OK", [("Content-Type", "text/plain")])
        response.body = b"{}"
        with pytest.raises(AssertionError):
            response.parse_json_body()

    def test_parse_json_body__wrong_content_encoding(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "application/json; charset=latin1")]
        )
        response.body = b"{}"
        with pytest.raises(AssertionError):
            response.parse_json_body()

    def test_parse_json_body__invalid_json(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "application/json")]
        )
        response.body = b'{"foo":'
        with pytest.raises(AssertionError):
            response.parse_json_body()

    def test_parse_json_body__invalid_encoding(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "application/json; charset=utf-8")]
        )
        response.body = '{"föo": 5}'.encode("iso-8859-1")
        with pytest.raises(AssertionError):
            response.parse_json_body()

    def test_assert_status__ok(self) -> None:
        response = FakeResponse("404 Not Found", [])
        response.assert_status(HTTPStatus.NOT_FOUND)

    def test_assert_status__fail(self) -> None:
        response = FakeResponse("404 Not Found", [])
        with pytest.raises(AssertionError):
            response.assert_status(HTTPStatus.OK)

    def test_assert_header_not_set__is_set(self) -> None:
        response = FakeResponse("200 OK", [("X-Foo", "value")])
        with pytest.raises(AssertionError):
            response.assert_header_not_set("x-FOO")

    def test_assert_header_not_set__not_set(self) -> None:
        response = FakeResponse("200 OK", [])
        response.assert_header_not_set("X-Foo")

    def test_assert_header_equal__no_such_header(self) -> None:
        response = FakeResponse("200 OK", [("X-Other", "value")])
        with pytest.raises(AssertionError):
            response.assert_header_equal("X-Header", "value")

    def test_assert_header_equal__ok(self) -> None:
        response = FakeResponse("200 OK", [("X-Header", "value")])
        response.assert_header_equal("X-Header", "value")

    def test_assert_header_equal__differs(self) -> None:
        response = FakeResponse("200 OK", [("X-Header", "other")])
        with pytest.raises(AssertionError):
            response.assert_header_equal("X-Header", "value")

    def test_assert_created_at__ok(self) -> None:
        response = FakeResponse(
            "201 Created", [("Location", "http://example.com/")]
        )
        response.assert_created_at("http://example.com/")

    def test_assert_created_at__wrong_status(self) -> None:
        response = FakeResponse(
            "200 OK", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_created_at("http://example.com/")

    def test_assert_created_at__no_location_header(self) -> None:
        response = FakeResponse("201 Created", [])
        with pytest.raises(AssertionError):
            response.assert_created_at("http://example.org/")

    def test_assert_created_at__wrong_location(self) -> None:
        response = FakeResponse(
            "201 Created", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_created_at("http://example.org/")

    def test_assert_created_at__relative_location(self) -> None:
        response = FakeResponse(
            "201 Created", [("Location", "http://example.com/foo/bar")]
        )
        response.assert_created_at("/foo/bar")

    def test_assert_created_at__keep_query_string(self) -> None:
        response = FakeResponse(
            "201 Created",
            [("Location", "http://example.com/foo?abc=def#frag")],
        )
        response.assert_created_at("/foo?abc=def#frag")

    def test_assert_see_other__ok(self) -> None:
        response = FakeResponse(
            "303 See Other", [("Location", "http://example.com/")]
        )
        response.assert_see_other("http://example.com/")

    def test_assert_see_other__wrong_status(self) -> None:
        response = FakeResponse(
            "200 OK", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_see_other("http://example.com/")

    def test_assert_see_other__no_location_header(self) -> None:
        response = FakeResponse("303 See Other", [])
        with pytest.raises(AssertionError):
            response.assert_see_other("http://example.org/")

    def test_assert_see_other__wrong_location(self) -> None:
        response = FakeResponse(
            "303 See Other", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_see_other("http://example.org/")

    def test_assert_see_other__relative_location(self) -> None:
        response = FakeResponse(
            "303 See Other", [("Location", "http://example.com/foo/bar")]
        )
        response.assert_see_other("/foo/bar")

    def test_assert_see_other__keep_query_string(self) -> None:
        response = FakeResponse(
            "303 See Other",
            [("Location", "http://example.com/foo?abc=def#frag")],
        )
        response.assert_see_other("/foo?abc=def#frag")

    def test_assert_temporary_redirect__ok(self) -> None:
        response = FakeResponse(
            "307 Temporary Redirect", [("Location", "http://example.com/")]
        )
        response.assert_temporary_redirect("http://example.com/")

    def test_assert_temporary_redirect__wrong_status(self) -> None:
        response = FakeResponse(
            "200 OK", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_temporary_redirect("http://example.com/")

    def test_assert_temporary_redirect__no_location_header(self) -> None:
        response = FakeResponse("307 Temporary Redirect", [])
        with pytest.raises(AssertionError):
            response.assert_temporary_redirect("http://example.org/")

    def test_assert_temporary_redirect__wrong_location(self) -> None:
        response = FakeResponse(
            "307 Temporary Redirect", [("Location", "http://example.com/")]
        )
        with pytest.raises(AssertionError):
            response.assert_temporary_redirect("http://example.org/")

    def test_assert_temporary_redirect__relative_location(self) -> None:
        response = FakeResponse(
            "307 Temporary Redirect",
            [("Location", "http://example.com/foo/bar")],
        )
        response.assert_temporary_redirect("/foo/bar")

    def test_assert_temporary_redirect__keep_query_string(self) -> None:
        response = FakeResponse(
            "307 Temporary Redirect",
            [("Location", "http://example.com/foo?abc=def#frag")],
        )
        response.assert_temporary_redirect("/foo?abc=def#frag")

    def test_assert_content_type__no_such_header(self) -> None:
        response = FakeResponse("200 OK", [])
        with pytest.raises(AssertionError):
            response.assert_content_type("image/png")

    def test_assert_content_type__equal(self) -> None:
        response = FakeResponse("200 OK", [("Content-Type", "image/png")])
        response.assert_content_type("image/png")

    def test_assert_content_type__different(self) -> None:
        response = FakeResponse("200 OK", [("Content-Type", "image/png")])
        with pytest.raises(AssertionError):
            response.assert_content_type("image/jpeg")

    def test_assert_content_type__charset_matches(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "text/html; charset=us-ascii")]
        )
        response.assert_content_type("text/html", charset="us-ascii")

    def test_assert_content_type__charset_list_matches(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "text/html; charset=us-ascii")]
        )
        response.assert_content_type(
            "text/html", charset=["us-ascii", "utf-8", None]
        )

    def test_assert_content_type__charset_list_matches__none(self) -> None:
        response = FakeResponse("200 OK", [("Content-Type", "text/html")])
        response.assert_content_type(
            "text/html", charset=["us-ascii", "utf-8", None]
        )

    def test_assert_content_type__charset_not_checked(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "text/html; charset=utf-8")]
        )
        response.assert_content_type("text/html")

    def test_assert_content_type__no_charset_in_response(self) -> None:
        response = FakeResponse("200 OK", [("Content-Type", "text/html")])
        with pytest.raises(AssertionError):
            response.assert_content_type("text/html", charset="us-ascii")

    def test_assert_content_type__wrong_charset(self) -> None:
        response = FakeResponse(
            "200 OK", [("Content-Type", "text/html; charset=utf-8")]
        )
        with pytest.raises(AssertionError):
            response.assert_content_type("text/html", charset="us-ascii")

    def test_assert_set_cookie__simple_match(self) -> None:
        response = FakeResponse(
            "200 OK", [("Set-Cookie", "Foo=Bar; Secure; Max-Age=1234")]
        )
        response.assert_set_cookie("Foo", "Bar")

    def test_assert_set_cookie__no_cookie_header(self) -> None:
        response = FakeResponse("200 OK", [])
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar")

    def test_assert_set_cookie__no_cookie_value(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo")])
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar")

    def test_assert_set_cookie__wrong_name(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Wrong=Bar")])
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar")

    def test_assert_set_cookie__wrong_value(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo=Wrong")])
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar")

    def test_assert_set_cookie__has_secure(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo=Bar; Secure")])
        response.assert_set_cookie("Foo", "Bar")
        response.assert_set_cookie("Foo", "Bar", secure=True)
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", secure=False)

    def test_assert_set_cookie__no_secure(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo=Bar")])
        response.assert_set_cookie("Foo", "Bar")
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", secure=True)
        response.assert_set_cookie("Foo", "Bar", secure=False)

    def test_assert_set_cookie__has_http_only(self) -> None:
        response = FakeResponse(
            "200 OK", [("Set-Cookie", "Foo=Bar; HttpOnly")]
        )
        response.assert_set_cookie("Foo", "Bar")
        response.assert_set_cookie("Foo", "Bar", http_only=True)
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", http_only=False)

    def test_assert_set_cookie__no_http_only(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo=Bar")])
        response.assert_set_cookie("Foo", "Bar")
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", http_only=True)
        response.assert_set_cookie("Foo", "Bar", http_only=False)

    def test_assert_set_cookie__has_max_age(self) -> None:
        response = FakeResponse(
            "200 OK", [("Set-Cookie", "Foo=Bar; Max-Age=1234")]
        )
        response.assert_set_cookie("Foo", "Bar")
        response.assert_set_cookie("Foo", "Bar", max_age=1234)
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", max_age=9999)

    def test_assert_set_cookie__invalid_max_age(self) -> None:
        response = FakeResponse(
            "200 OK", [("Set-Cookie", "Foo=Bar; Max-Age=INVALID")]
        )
        response.assert_set_cookie("Foo", "Bar")
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", max_age=1234)

    def test_assert_set_cookie__no_max_age(self) -> None:
        response = FakeResponse("200 OK", [("Set-Cookie", "Foo=Bar")])
        response.assert_set_cookie("Foo", "Bar")
        with pytest.raises(AssertionError):
            response.assert_set_cookie("Foo", "Bar", max_age=1234)


class TestWSGIAppTest:
    def test_run_app(self) -> None:
        app_run = False
        env: WSGIEnvironment | None = None

        def app(
            environ: WSGIEnvironment, sr: StartResponse
        ) -> Iterable[bytes]:
            nonlocal app_run, env
            app_run = True
            env = environ
            sr("200 OK", [])
            return []

        request = create_request("GET", "/foo/bar")
        request.add_argument("foo", "bar")
        run_wsgi_test(app, request)
        assert app_run, "app not run"
        assert env is not None
        assert env.get("QUERY_STRING") == "foo=bar"

    def test_response(self) -> None:
        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            sr("404 Not Found", [("X-Foo", "Bar")])
            return []

        request = create_request("GET", "/foo/bar")
        response = run_wsgi_test(app, request)
        response.assert_status(HTTPStatus.NOT_FOUND)
        response.assert_header_equal("X-Foo", "Bar")

    def test_response_body_without_close(self) -> None:
        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            writer = sr("200 OK", [])
            writer(b"Abc")
            writer(b"def")
            return [b"Foo", b"bar"]

        request = create_request("GET", "/foo/bar")
        response = run_wsgi_test(app, request)
        assert response.body == b"AbcdefFoobar"

    def test_response_body_with_close(self) -> None:
        file = BytesIO(b"Foo")

        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            writer = sr("200 OK", [])
            writer(b"Abc")
            writer(b"def")
            return file

        request = create_request("GET", "/foo/bar")
        response = run_wsgi_test(app, request)
        assert response.body == b"AbcdefFoo"
        assert file.closed

    def test_start_response_not_called(self) -> None:
        def app(_: WSGIEnvironment, __: StartResponse) -> Iterable[bytes]:
            return []

        request = create_request("GET", "/foo/bar")
        with pytest.raises(AssertionError):
            run_wsgi_test(app, request)

    def test_start_response_called_multiple_times(self) -> None:
        assert_raised = False

        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            nonlocal assert_raised
            sr("200 OK", [])
            try:
                sr("404 Not Found", [])
            except AssertionError:
                assert_raised = True
            return []

        request = create_request("GET", "/foo/bar")
        run_wsgi_test(app, request)
        assert assert_raised

    def test_start_response_called_multiple_times_with_exc_info(self) -> None:
        assert_raised = False

        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            nonlocal assert_raised
            sr("200 OK", [])
            try:
                sr("404 Not Found", [], _get_exc_info())
            except AssertionError:
                assert_raised = True
            return []

        request = create_request("GET", "/foo/bar")
        run_wsgi_test(app, request)
        assert not assert_raised

    def test_start_response_called_after_output_written(self) -> None:
        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            writer = sr("200 OK", [])
            writer(b"abc")
            sr("404 OK", [], _get_exc_info())
            return []

        request = create_request("GET", "/foo/bar")
        with pytest.raises(ValueError):
            run_wsgi_test(app, request)

    def test_start_response_called_no_output_written(self) -> None:
        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            writer = sr("200 OK", [])
            writer(b"")
            sr("404 OK", [], _get_exc_info())
            return []

        request = create_request("GET", "/foo/bar")
        response = run_wsgi_test(app, request)
        response.assert_status(HTTPStatus.NOT_FOUND)


def _get_exc_info() -> tuple[Any, Any, Any]:
    try:
        raise ValueError()
    except:  # noqa
        return sys.exc_info()


class TestWSGIArgumentsTest:
    def _create_app(
        self, argument_template: Sequence[ArgumentTemplate]
    ) -> WSGIApplication:
        def app(env: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            try:
                parse_args(env, argument_template)
            except ArgumentsError:
                sr("400 Bad Request", [])
            else:
                sr("200 OK", [])
            return []

        return app

    def _successful_arg_test(
        self,
        app_args: Sequence[ArgumentTemplate],
        expected_args: Iterable[ArgumentToTest],
    ) -> None:
        app = self._create_app(app_args)
        request = create_request("GET", "/")
        run_wsgi_arguments_test(app, request, expected_args)

    def _failing_arg_test(
        self,
        app_args: Sequence[ArgumentTemplate],
        expected_args: Iterable[ArgumentToTest],
    ) -> None:
        app = self._create_app(app_args)
        request = create_request("GET", "/")
        with pytest.raises(AssertionError):
            run_wsgi_arguments_test(app, request, expected_args)

    def test_no_expected_args(self) -> None:
        self._successful_arg_test([], [])

    def test_required_argument_present(self) -> None:
        self._successful_arg_test(
            [("arg", int, Multiplicity.REQUIRED)],
            [("arg", Multiplicity.REQUIRED, "42")],
        )

    def test_required_argument_not_in_app(self) -> None:
        self._failing_arg_test(
            [("arg", int, Multiplicity.OPTIONAL)],
            [("arg", Multiplicity.REQUIRED, "42")],
        )

    def test_required_argument_not_in_test(self) -> None:
        self._failing_arg_test([("arg", int, Multiplicity.REQUIRED)], [])

    def test_required_argument_optional_in_test(self) -> None:
        self._failing_arg_test(
            [("arg", int, Multiplicity.REQUIRED)],
            [("arg", Multiplicity.OPTIONAL, "42")],
        )

    def test_required_any_argument_present(self) -> None:
        self._successful_arg_test(
            [("arg", int, Multiplicity.REQUIRED_ANY)],
            [("arg", Multiplicity.REQUIRED_ANY, "42")],
        )

    def test_required_any_argument_not_in_app(self) -> None:
        self._failing_arg_test(
            [("arg", int, Multiplicity.OPTIONAL)],
            [("arg", Multiplicity.REQUIRED_ANY, "42")],
        )

    def test_required_any_argument_not_in_test(self) -> None:
        self._failing_arg_test([("arg", int, Multiplicity.REQUIRED_ANY)], [])

    def test_required_any_argument_optional_in_test(self) -> None:
        self._failing_arg_test(
            [("arg", int, Multiplicity.REQUIRED_ANY)],
            [("arg", Multiplicity.OPTIONAL, "42")],
        )

    def test_optional_argument_not_in_app(self) -> None:
        self._successful_arg_test([], [("arg", Multiplicity.OPTIONAL, "foo")])

    def test_optional_argument_not_in_test(self) -> None:
        self._successful_arg_test([("arg", int, Multiplicity.OPTIONAL)], [])

    def test_any_argument_not_in_app(self) -> None:
        self._successful_arg_test([], [("arg", Multiplicity.ANY, "foo")])

    def test_any_argument_not_in_test(self) -> None:
        self._successful_arg_test([("arg", int, Multiplicity.ANY)], [])

    def test_correct_value_not_accepted(self) -> None:
        self._failing_arg_test(
            [("arg", int, Multiplicity.OPTIONAL)],
            [("arg", Multiplicity.OPTIONAL, "not-a-number")],
        )

    def test_invalid_value_accepted(self) -> None:
        self._failing_arg_test(
            [("arg", str, Multiplicity.OPTIONAL)],
            [("arg", Multiplicity.OPTIONAL, "42", "not-a-number")],
        )

    def test_handle_other_errors(self) -> None:
        def app(_: WSGIEnvironment, sr: StartResponse) -> Iterable[bytes]:
            sr("500 Internal Server Error", [])
            return []

        request = create_request("POST", "/")
        with pytest.raises(AssertionError):
            run_wsgi_arguments_test(app, request, [])

    def test_post_request__no_args(self) -> None:
        app = self._create_app([("arg", int, Multiplicity.OPTIONAL)])
        request = create_request("POST", "/")
        run_wsgi_arguments_test(app, request, [])
